"""Test traffic rules API for disabling and enabling traffic rules.

pytest --cov-report term-missing --cov=aiounifi.traffic_rule tests/test_traffic_rules.py
"""

import pytest

from aiounifi.models.traffic_rule import TrafficRuleEnableRequest

from .fixtures import TRAFFIC_RULES, WIRELESS_CLIENT


@pytest.mark.parametrize("is_unifi_os", [True])
@pytest.mark.parametrize("enable", [True, False])
async def test_traffic_rule_enable_request(
    mock_aioresponse, unifi_controller, unifi_called_with, enable
):
    """Test that traffic rule can be enabled and disabled."""
    traffic_rule_enabled = TRAFFIC_RULES[0]
    traffic_rule_enabled_id = traffic_rule_enabled["_id"]
    traffic_rule_disabled = TRAFFIC_RULES[1]
    traffic_rule_disabled_id = traffic_rule_disabled["_id"]

    traffic_rule_id = traffic_rule_disabled_id if enable else traffic_rule_enabled_id
    traffic_rule = traffic_rule_disabled if enable else traffic_rule_enabled
    mock_aioresponse.put(
        "https://host:8443/proxy/network/v2/api/site/default"
        + f"/trafficrules/{traffic_rule_id}",
        payload={},
    )

    await unifi_controller.request(
        TrafficRuleEnableRequest.create(traffic_rule, enable)
    )

    traffic_rule["enabled"] = enable
    assert unifi_called_with(
        "put",
        f"/proxy/network/v2/api/site/default/trafficrules/{traffic_rule_id}",
        json=traffic_rule,
    )


@pytest.mark.parametrize("traffic_rule_payload", [TRAFFIC_RULES])
@pytest.mark.parametrize("is_unifi_os", [True])
@pytest.mark.parametrize("enable", [True, False])
async def test_traffic_rule_toggle(
    mock_aioresponse, unifi_controller, _mock_endpoints, enable
):
    """Test toggle method can enable and disable a traffic rule."""
    traffic_rules = unifi_controller.traffic_rules
    await traffic_rules.update()

    traffic_rule_id = TRAFFIC_RULES[0 if not enable else 1]["_id"]

    mock_aioresponse.put(
        "https://host:8443/proxy/network/v2/api/site/default"
        + f"/trafficrules/{traffic_rule_id}",
        payload={},
    )
    await traffic_rules.toggle(traffic_rules[traffic_rule_id], enable)
    assert traffic_rules[traffic_rule_id].enabled is enable


@pytest.mark.parametrize("traffic_rule_payload", [TRAFFIC_RULES])
@pytest.mark.parametrize("is_unifi_os", [True])
@pytest.mark.parametrize("enable", [True, False])
async def test_traffic_rule_enable_disable(
    mock_aioresponse, unifi_controller, _mock_endpoints, enable
):
    """Test individual methods for enabled and disabled."""
    traffic_rules = unifi_controller.traffic_rules
    await traffic_rules.update()

    traffic_rule_id = TRAFFIC_RULES[0 if not enable else 1]["_id"]
    traffic_rule_call = traffic_rules.disable if not enable else traffic_rules.enable

    mock_aioresponse.put(
        "https://host:8443/proxy/network/v2/api/site/default"
        + f"/trafficrules/{traffic_rule_id}",
        payload={},
    )
    await traffic_rule_call(traffic_rules[traffic_rule_id])
    assert traffic_rules[traffic_rule_id].enabled is enable


@pytest.mark.parametrize("is_unifi_os", [True])
async def test_no_traffic_rules(unifi_controller, _mock_endpoints, unifi_called_with):
    """Test that no traffic rules also work."""
    traffic_rules = unifi_controller.traffic_rules
    await traffic_rules.update()
    assert unifi_called_with("get", "/proxy/network/v2/api/site/default/trafficrules")
    assert len(traffic_rules.values()) == 0


@pytest.mark.parametrize("traffic_rule_payload", [TRAFFIC_RULES])
async def test_traffic_rules(unifi_controller, _mock_endpoints, unifi_called_with):
    """Test that we get the expected traffic rule."""
    traffic_rules = unifi_controller.traffic_rules
    await traffic_rules.update()
    assert unifi_called_with("get", "/v2/api/site/default/trafficrules")
    assert len(traffic_rules.values()) == 2

    traffic_rule = traffic_rules["6452cd9b859d5b11aa002ea1"]
    assert traffic_rule.id == "6452cd9b859d5b11aa002ea1"
    assert traffic_rule.description == "Test 1"
    assert traffic_rule.enabled is False
    assert traffic_rule.action == "BLOCK"
    assert traffic_rule.matching_target == "INTERNET"
    assert traffic_rule.target_devices == [
        {"client_mac": WIRELESS_CLIENT["mac"], "type": "CLIENT"}
    ]



"""Traffic rules as part of a UniFi network."""

from copy import deepcopy

from ..models.api import TypedApiResponse
from ..models.traffic_rule import (
    TrafficRule,
    TrafficRuleEnableRequest,
    TrafficRuleListRequest,
)
from .api_handlers import APIHandler


class TrafficRules(APIHandler[TrafficRule]):
    """Represents TrafficRules configurations."""

    obj_id_key = "_id"
    item_cls = TrafficRule
    api_request = TrafficRuleListRequest.create()

    async def enable(self, traffic_rule: TrafficRule) -> TypedApiResponse:
        """Enable traffic rule defined in controller."""
        return await self.toggle(traffic_rule, state=True)

    async def disable(self, traffic_rule: TrafficRule) -> TypedApiResponse:
        """Disable traffic rule defined in controller."""
        return await self.toggle(traffic_rule, state=False)

    async def toggle(self, traffic_rule: TrafficRule, state: bool) -> TypedApiResponse:
        """Set traffic rule - defined in controller - to the desired state."""
        traffic_rule_dict = deepcopy(traffic_rule.raw)
        traffic_rule_response = await self.controller.request(
            TrafficRuleEnableRequest.create(traffic_rule_dict, enable=state)
        )
        self.process_raw(traffic_rule_response.get("data", []))
        return traffic_rule_response



"""Python library to interact with UniFi controller."""

import asyncio
from collections.abc import Callable, Coroutine
import logging
from typing import TYPE_CHECKING, Any

from .interfaces.clients import Clients
from .interfaces.clients_all import ClientsAll
from .interfaces.connectivity import Connectivity
from .interfaces.devices import Devices
from .interfaces.dpi_restriction_apps import DPIRestrictionApps
from .interfaces.dpi_restriction_groups import DPIRestrictionGroups
from .interfaces.events import EventHandler
from .interfaces.messages import MessageHandler
from .interfaces.outlets import Outlets
from .interfaces.port_forwarding import PortForwarding
from .interfaces.ports import Ports
from .interfaces.sites import Sites
from .interfaces.system_information import SystemInformationHandler
from .interfaces.traffic_routes import TrafficRoutes
from .interfaces.traffic_rules import TrafficRules
from .interfaces.wlans import Wlans
from .models.configuration import Configuration

if TYPE_CHECKING:
    from .models.api import ApiRequest, TypedApiResponse

LOGGER = logging.getLogger(__name__)


class Controller:
    """Control a UniFi controller."""

    def __init__(self, config: Configuration) -> None:
        """Session setup."""
        self.connectivity = Connectivity(config)

        self.messages = MessageHandler(self)
        self.events = EventHandler(self)

        self.clients = Clients(self)
        self.clients_all = ClientsAll(self)
        self.devices = Devices(self)
        self.outlets = Outlets(self)
        self.ports = Ports(self)
        self.dpi_apps = DPIRestrictionApps(self)
        self.dpi_groups = DPIRestrictionGroups(self)
        self.port_forwarding = PortForwarding(self)
        self.sites = Sites(self)
        self.system_information = SystemInformationHandler(self)
        self.traffic_rules = TrafficRules(self)
        self.traffic_routes = TrafficRoutes(self)
        self.wlans = Wlans(self)

        self.update_handlers: tuple[Callable[[], Coroutine[Any, Any, None]], ...] = (
            self.clients.update,
            self.clients_all.update,
            self.devices.update,
            self.dpi_apps.update,
            self.dpi_groups.update,
            self.port_forwarding.update,
            self.sites.update,
            self.system_information.update,
            self.traffic_rules.update,
            self.traffic_routes.update,
            self.wlans.update,
        )

    async def login(self) -> None:
        """Log in to controller."""
        await self.connectivity.check_unifi_os()
        await self.connectivity.login()

    async def request(self, api_request: "ApiRequest") -> "TypedApiResponse":
        """Make a request to the API, retry login on failure."""
        return await self.connectivity.request(api_request)

    async def initialize(self) -> None:
        """Load UniFi parameters."""
        results = await asyncio.gather(
            *[update() for update in self.update_handlers],
            return_exceptions=True,
        )
        for result in results:
            if result is not None:
                LOGGER.warning("Exception on update %s", result)

    async def start_websocket(self) -> None:
        """Start websocket session."""
        await self.connectivity.websocket(self.messages.new_data)



"""API management class and base class for the different end points."""

from abc import ABC
from collections.abc import Callable, ItemsView, Iterator, ValuesView
import enum
from typing import TYPE_CHECKING, Any, Generic, final

from ..models.api import ApiItemT, ApiRequest

if TYPE_CHECKING:
    from ..controller import Controller
    from ..models.message import Message, MessageKey


class ItemEvent(enum.Enum):
    """The event action of the item."""

    ADDED = "added"
    CHANGED = "changed"
    DELETED = "deleted"


CallbackType = Callable[[ItemEvent, str], None]
SubscriptionType = tuple[CallbackType, tuple[ItemEvent, ...] | None]
UnsubscribeType = Callable[[], None]

ID_FILTER_ALL = "*"


class SubscriptionHandler(ABC):
    """Manage subscription and notification to subscribers."""

    def __init__(self) -> None:
        """Initialize subscription handler."""
        self._subscribers: dict[str, list[SubscriptionType]] = {ID_FILTER_ALL: []}

    def signal_subscribers(self, event: ItemEvent, obj_id: str) -> None:
        """Signal subscribers."""
        subscribers: list[SubscriptionType] = (
            self._subscribers.get(obj_id, []) + self._subscribers[ID_FILTER_ALL]
        )
        for callback, event_filter in subscribers:
            if event_filter is not None and event not in event_filter:
                continue
            callback(event, obj_id)

    def subscribe(
        self,
        callback: CallbackType,
        event_filter: tuple[ItemEvent, ...] | ItemEvent | None = None,
        id_filter: tuple[str] | str | None = None,
    ) -> UnsubscribeType:
        """Subscribe to added events."""
        if isinstance(event_filter, ItemEvent):
            event_filter = (event_filter,)
        subscription = (callback, event_filter)

        _id_filter: tuple[str]
        if id_filter is None:
            _id_filter = (ID_FILTER_ALL,)
        elif isinstance(id_filter, str):
            _id_filter = (id_filter,)

        for obj_id in _id_filter:
            if obj_id not in self._subscribers:
                self._subscribers[obj_id] = []
            self._subscribers[obj_id].append(subscription)

        def unsubscribe() -> None:
            for obj_id in _id_filter:
                if obj_id not in self._subscribers:
                    continue
                if subscription not in self._subscribers[obj_id]:
                    continue
                self._subscribers[obj_id].remove(subscription)

        return unsubscribe


class APIHandler(SubscriptionHandler, Generic[ApiItemT]):
    """Base class for a map of API Items."""

    obj_id_key: str
    item_cls: type[ApiItemT]
    api_request: ApiRequest
    process_messages: tuple["MessageKey", ...] = ()
    remove_messages: tuple["MessageKey", ...] = ()

    def __init__(self, controller: "Controller") -> None:
        """Initialize API handler."""
        super().__init__()
        self.controller = controller
        self._items: dict[str, ApiItemT] = {}

        if message_filter := self.process_messages + self.remove_messages:
            controller.messages.subscribe(self.process_message, message_filter)

    @final
    async def update(self) -> None:
        """Refresh data."""
        raw = await self.controller.request(self.api_request)
        self.process_raw(raw.get("data", []))

    @final
    def process_raw(self, raw: list[dict[str, Any]]) -> None:
        """Process full raw response."""
        for raw_item in raw:
            self.process_item(raw_item)

    @final
    def process_message(self, message: "Message") -> None:
        """Process and forward websocket data."""
        if message.meta.message in self.process_messages:
            self.process_item(message.data)

        elif message.meta.message in self.remove_messages:
            self.remove_item(message.data)

    @final
    def process_item(self, raw: dict[str, Any]) -> None:
        """Process item data."""
        if self.obj_id_key not in raw:
            return

        obj_id: str
        obj_is_known = (obj_id := raw[self.obj_id_key]) in self._items
        self._items[obj_id] = self.item_cls(raw)

        self.signal_subscribers(
            ItemEvent.CHANGED if obj_is_known else ItemEvent.ADDED,
            obj_id,
        )

    @final
    def remove_item(self, raw: dict[str, Any]) -> None:
        """Remove item."""
        obj_id: str
        if (obj_id := raw[self.obj_id_key]) in self._items:
            self._items.pop(obj_id)
            self.signal_subscribers(ItemEvent.DELETED, obj_id)

    @final
    def items(self) -> ItemsView[str, ApiItemT]:
        """Return items dictionary."""
        return self._items.items()

    @final
    def values(self) -> ValuesView[ApiItemT]:
        """Return items."""
        return self._items.values()

    @final
    def get(self, obj_id: str, default: Any | None = None) -> ApiItemT | None:
        """Get item value based on key, return default if no match."""
        return self._items.get(obj_id, default)

    @final
    def __contains__(self, obj_id: str) -> bool:
        """Validate membership of item ID."""
        return obj_id in self._items

    @final
    def __getitem__(self, obj_id: str) -> ApiItemT:
        """Get item value based on key."""
        return self._items[obj_id]

    @final
    def __iter__(self) -> Iterator[str]:
        """Allow iterate over items."""
        return iter(self._items)



"""Device port handler."""

from typing import TYPE_CHECKING

from ..models.port import Port
from .api_handlers import APIHandler, ItemEvent

if TYPE_CHECKING:
    from ..controller import Controller


class Ports(APIHandler[Port]):
    """Represents network device ports."""

    item_cls = Port

    def __init__(self, controller: "Controller") -> None:
        """Initialize API handler."""
        super().__init__(controller)
        controller.devices.subscribe(self.process_device)

    def process_device(self, event: ItemEvent, device_id: str) -> None:
        """Add, update, remove."""
        if event in (ItemEvent.ADDED, ItemEvent.CHANGED):
            device = self.controller.devices[device_id]
            if "port_table" not in device.raw:
                return
            for raw_port in device.raw["port_table"]:
                port = Port(raw_port)
                if (port_idx := port.port_idx or port.ifname) is None:
                    continue
                obj_id = f"{device_id}_{port_idx}"
                self._items[obj_id] = port
                self.signal_subscribers(event, obj_id)
            return

        matched_obj_ids = [
            obj_id for obj_id in self._items if obj_id.startswith(device_id)
        ]
        for obj_id in matched_obj_ids:
            self._items.pop(obj_id)
            self.signal_subscribers(event, obj_id)



"""UniFi port forwarding."""

from ..models.message import MessageKey
from ..models.port_forward import PortForward, PortForwardListRequest
from .api_handlers import APIHandler


class PortForwarding(APIHandler[PortForward]):
    """Represents port forwarding."""

    obj_id_key = "_id"
    item_cls = PortForward
    process_messages = (MessageKey.PORT_FORWARD_ADDED, MessageKey.PORT_FORWARD_UPDATED)
    remove_messages = (MessageKey.PORT_FORWARD_DELETED,)
    api_request = PortForwardListRequest.create()



"""Test aioUniFi device ports.

pytest --cov-report term-missing --cov=aiounifi.ports tests/test_ports.py
"""
from unittest.mock import Mock

import pytest

from aiounifi.interfaces.api_handlers import ItemEvent
from aiounifi.models.port import Port

from .fixtures import SWITCH_16_PORT_POE


async def test_handler_ports(unifi_controller):
    """Verify that device ports works."""
    ports = unifi_controller.ports
    assert len(ports.items()) == 0
    assert len(ports.values()) == 0

    assert "fc:ec:da:11:22:33_18" not in ports
    assert ports.get("fc:ec:da:11:22:33_18", "xy") == "xy"
    assert not isinstance(ports.get("fc:ec:da:11:22:33_18"), Port)
    unsub_all = ports.subscribe(mock_subscribe_all := Mock())
    unsub_18 = ports.subscribe(
        mock_subscribe_18 := Mock(), id_filter="fc:ec:da:11:22:33_18"
    )
    unsub_added = ports.subscribe(
        mock_subscribe_added := Mock(), event_filter=ItemEvent.ADDED
    )
    unsub_changed = ports.subscribe(
        mock_subscribe_changed := Mock(), event_filter=ItemEvent.CHANGED
    )
    unsub_deleted = ports.subscribe(
        mock_subscribe_deleted := Mock(), event_filter=ItemEvent.DELETED
    )
    unsub_bad = ports.subscribe(mock_subscribe_bad := Mock(), id_filter="bad")

    # Add ports
    unifi_controller.devices.process_raw([SWITCH_16_PORT_POE])
    assert next(iter(ports)) == "fc:ec:da:11:22:33_1"
    assert "fc:ec:da:11:22:33_18" in ports
    assert isinstance(ports.get("fc:ec:da:11:22:33_18"), Port)
    assert isinstance(ports["fc:ec:da:11:22:33_18"], Port)
    assert len(ports.values()) == 18
    assert mock_subscribe_all.call_count == 18
    mock_subscribe_all.assert_called_with(ItemEvent.ADDED, "fc:ec:da:11:22:33_18")
    mock_subscribe_18.assert_called_once()
    assert mock_subscribe_added.call_count == 18
    mock_subscribe_changed.assert_not_called()
    mock_subscribe_deleted.assert_not_called()
    mock_subscribe_bad.assert_not_called()

    # Update ports
    unifi_controller.devices.process_raw([SWITCH_16_PORT_POE])
    assert len(ports.values()) == 18
    assert mock_subscribe_all.call_count == 36
    mock_subscribe_all.assert_called_with(ItemEvent.CHANGED, "fc:ec:da:11:22:33_18")
    assert mock_subscribe_18.call_count == 2
    assert mock_subscribe_added.call_count == 18
    assert mock_subscribe_changed.call_count == 18
    mock_subscribe_deleted.assert_not_called()
    mock_subscribe_bad.assert_not_called()

    # Remove ports
    unifi_controller.devices.remove_item(SWITCH_16_PORT_POE)
    assert len(ports.values()) == 0
    assert mock_subscribe_all.call_count == 54
    mock_subscribe_all.assert_called_with(ItemEvent.DELETED, "fc:ec:da:11:22:33_18")
    assert mock_subscribe_18.call_count == 3
    assert mock_subscribe_added.call_count == 18
    assert mock_subscribe_changed.call_count == 18
    assert mock_subscribe_deleted.call_count == 18
    mock_subscribe_bad.assert_not_called()

    unsub_all()
    unsub_18()
    unsub_added()
    unsub_changed()
    unsub_deleted()
    unsub_bad()

    unsub_all()
    unsub_all()


async def test_handler_process_device_no_index(unifi_controller):
    """Verify that device ports works."""
    ports = unifi_controller.ports
    unifi_controller.devices.process_raw([{"mac": "1", "port_table": [{}]}])
    assert len(ports.items()) == 0


@pytest.mark.parametrize("device_payload", [[SWITCH_16_PORT_POE]])
async def test_port(unifi_controller, _mock_endpoints):
    """Verify that device port model works."""
    await unifi_controller.devices.update()
    port = unifi_controller.ports["fc:ec:da:11:22:33_1"]

    assert port.ifname is None
    assert port.media == "GE"
    assert port.name == "Port 1"
    assert port.port_idx == 1
    assert port.poe_class == "Unknown"
    assert port.poe_caps == 7
    assert port.poe_enable is False
    assert port.poe_mode == "auto"
    assert port.poe_power == "0.00"
    assert port.poe_voltage == "0.00"
    assert port.portconf_id == "5a32aa4ee4babd4452422ddd22222"
    assert port.port_poe is True
    assert port.up is True

    assert repr(port) == "<Port 1: Poe False>"



"""Test client configuration API.

pytest --cov-report term-missing --cov=aiounifi.devices tests/test_devices.py
"""
from collections.abc import Callable
from typing import Any
from unittest.mock import Mock

from aioresponses import aioresponses
import pytest

from aiounifi.controller import Controller
from aiounifi.models.device import (
    DevicePowerCyclePortRequest,
    DeviceRestartRequest,
    DeviceSetOutletCycleEnabledRequest,
    DeviceSetOutletRelayRequest,
    DeviceSetPoePortModeRequest,
    DeviceState,
    DeviceUpgradeRequest,
)
from aiounifi.models.message import MessageKey

from .fixtures import (
    ACCESS_POINT_AC_PRO,
    GATEWAY_USG3,
    PDU_PRO,
    PLUG_UP1,
    STRIP_UP6,
    SWITCH_16_PORT_POE,
)

test_data = [
    ([{"mac": "0", "disabled": True}], {"mac": "0", "disabled": True}),
    (
        [ACCESS_POINT_AC_PRO],
        {
            "board_revision": 21,
            "considered_lost_at": 1588175837,
            "disabled": False,
            "id": "235678987654345678",
            "ip": "192.168.0.4",
            "downlink_table": [],
            "fan_level": None,
            "has_fan": False,
            "last_seen": 1588175726,
            "lldp_table": [],
            "mac": "80:2a:a8:00:01:02",
            "model": "U7PG2",
            "name": "ACCESS POINT AC PRO",
            "next_heartbeat_at": 1588175763,
            "overheating": False,
            "port_overrides": [],
            "port_table": ACCESS_POINT_AC_PRO["port_table"],
            "speedtest_status": None,
            "state": 1,
            "sys_stats": {
                "loadavg_1": "0.15",
                "loadavg_15": "0.02",
                "loadavg_5": "0.08",
                "mem_buffer": 0,
                "mem_total": 128622592,
                "mem_used": 63606784,
            },
            "type": "uap",
            "version": "4.0.69.10871",
            "upgradable": True,
            "upgrade_to_firmware": "4.0.80.10875",
            "uplink": ACCESS_POINT_AC_PRO["uplink"],
            "uplink_depth": None,
            "user_num_sta": 12,
            "wlan_overrides": [
                {
                    "name": "My5GHzSSID1",
                    "radio": "na",
                    "radio_name": "wifi1",
                    "wlan_id": "012345678910111213141516",
                },
            ],
        },
    ),
    (
        [GATEWAY_USG3],
        {
            "board_revision": 16,
            "considered_lost_at": 1588175842,
            "disabled": False,
            "general_temperature": None,
            "has_temperature": False,
            "id": "235678987654345678",
            "ip": "1.2.3.4",
            "downlink_table": [],
            "fan_level": None,
            "has_fan": False,
            "last_seen": 1588175740,
            "lldp_table": [],
            "mac": "78:8a:20:33:44:55",
            "model": "UGW3",
            "name": "USG",
            "next_heartbeat_at": 1588175774,
            "overheating": False,
            "port_overrides": [],
            "port_table": GATEWAY_USG3["port_table"],
            "state": 1,
            "sys_stats": {
                "loadavg_1": "0.03",
                "loadavg_15": "0.08",
                "loadavg_5": "0.07",
                "mem_buffer": 57561088,
                "mem_total": 507412480,
                "mem_used": 293453824,
            },
            "system_stats": ("0", "25", "3971505"),
            "type": "ugw",
            "version": "4.4.44.5213844",
            "upgradable": True,
            "upgrade_to_firmware": "4.4.50.5272448",
            "uplink": GATEWAY_USG3["uplink"],
            "uplink_depth": None,
            "uptime": 3971869,
            "user_num_sta": 20,
            "wlan_overrides": [],
            "speedtest_status": GATEWAY_USG3["speedtest-status"],
        },
    ),
    (
        [PLUG_UP1],
        {
            "board_revision": 2,
            "downlink_table": [],
            "id": "600c8356942a6ade50707b56",
            "ip": "192.168.0.189",
            "has_fan": False,
            "has_temperature": False,
            "general_temperature": None,
            "last_seen": 1642055273,
            "lldp_table": [],
            "mac": "fc:ec:da:76:4f:5f",
            "model": "UP1",
            "name": "Plug",
            "next_interval": 40,
            "outlet_table": [
                {
                    "index": 1,
                    "has_relay": True,
                    "has_metering": False,
                    "relay_state": False,
                    "name": "Outlet 1",
                }
            ],
            "outlet_overrides": [],
            "port_table": [],
            "state": 1,
            "sys_stats": {"mem_total": 98304, "mem_used": 87736},
            "system_stats": ("", "", ""),
            "type": "uap",
            "version": "2.2.1.511",
            "upgradable": False,
            "uplink": PLUG_UP1["uplink"],
            "uptime": 376083,
        },
    ),
    (
        [STRIP_UP6],
        {
            "board_revision": 5,
            "downlink_table": [],
            "general_temperature": None,
            "has_temperature": False,
            "id": "61eb1a75942a6a859b45d2bc",
            "ip": "192.168.0.138",
            "has_fan": False,
            "last_seen": 1642800247,
            "lldp_table": [],
            "mac": "78:45:58:fc:16:7d",
            "model": "UP6",
            "name": "",
            "next_interval": 41,
            "outlet_table": [
                {
                    "index": 1,
                    "has_relay": True,
                    "has_metering": False,
                    "relay_state": False,
                    "cycle_enabled": False,
                    "name": "Outlet 1",
                },
                {
                    "index": 2,
                    "has_relay": True,
                    "has_metering": False,
                    "relay_state": False,
                    "cycle_enabled": False,
                    "name": "Outlet 2",
                },
                {
                    "index": 3,
                    "has_relay": True,
                    "has_metering": False,
                    "relay_state": True,
                    "cycle_enabled": False,
                    "name": "Outlet 3",
                },
                {
                    "index": 4,
                    "has_relay": True,
                    "has_metering": False,
                    "relay_state": True,
                    "cycle_enabled": True,
                    "name": "Outlet 4",
                },
                {
                    "index": 5,
                    "has_relay": True,
                    "has_metering": False,
                    "relay_state": False,
                    "cycle_enabled": False,
                    "name": "Outlet 5",
                },
                {
                    "index": 6,
                    "has_relay": True,
                    "has_metering": False,
                    "relay_state": False,
                    "cycle_enabled": False,
                    "name": "Outlet 6",
                },
                {
                    "index": 7,
                    "has_relay": True,
                    "has_metering": False,
                    "relay_state": False,
                    "cycle_enabled": False,
                    "name": "USB Outlets",
                },
            ],
            "outlet_overrides": [
                {
                    "index": 1,
                    "name": "Outlet 1",
                    "cycle_enabled": False,
                    "relay_state": False,
                },
                {
                    "index": 2,
                    "name": "Outlet 2",
                    "cycle_enabled": False,
                    "relay_state": False,
                },
                {
                    "index": 3,
                    "name": "Outlet 3",
                    "cycle_enabled": False,
                    "relay_state": True,
                },
                {
                    "index": 4,
                    "name": "Outlet 4",
                    "cycle_enabled": True,
                    "relay_state": True,
                },
                {
                    "index": 5,
                    "name": "Outlet 5",
                    "cycle_enabled": False,
                    "relay_state": False,
                },
                {
                    "index": 6,
                    "name": "Outlet 6",
                    "cycle_enabled": False,
                    "relay_state": False,
                },
                {
                    "index": 7,
                    "name": "USB Outlets",
                    "cycle_enabled": False,
                    "relay_state": False,
                },
            ],
            "port_table": [],
            "state": 1,
            "sys_stats": {"mem_total": 98304, "mem_used": 88056},
            "system_stats": ("", "", ""),
            "type": "uap",
            "version": "2.2.1.511",
            "upgradable": False,
            "uplink": STRIP_UP6["uplink"],
            "uptime": 2370,
        },
    ),
    (
        [PDU_PRO],
        {
            "board_revision": 1,
            "downlink_table": [],
            "general_temperature": None,
            "has_temperature": False,
            "id": "61e4a1e60bbb2d53aeb430ea",
            "ip": "192.168.1.66",
            "has_fan": False,
            "last_seen": 1643721168,
            "lldp_table": [
                {
                    "chassis_id": "00:00:00:00:00:83",
                    "chassis_id_subtype": "mac",
                    "is_wired": True,
                    "local_port_idx": 1,
                    "local_port_name": "eth0",
                    "port_id": "local Port 1",
                }
            ],
            "mac": "00:00:00:00:00:84",
            "model": "USPPDUP",
            "name": "Main Server Cabinet PDU",
            "next_interval": 56,
            "outlet_ac_power_budget": "1875.000",
            "outlet_ac_power_consumption": "307.741",
            "outlet_table": [
                {
                    "index": 1,
                    "relay_state": True,
                    "cycle_enabled": False,
                    "name": "USB Outlet 1",
                    "outlet_caps": 1,
                },
                {
                    "index": 2,
                    "relay_state": False,
                    "cycle_enabled": False,
                    "name": "USB Outlet 2",
                    "outlet_caps": 1,
                },
                {
                    "index": 3,
                    "relay_state": True,
                    "cycle_enabled": False,
                    "name": "USB Outlet 3",
                    "outlet_caps": 1,
                },
                {
                    "index": 4,
                    "relay_state": False,
                    "cycle_enabled": False,
                    "name": "USB Outlet 4",
                    "outlet_caps": 1,
                },
                {
                    "index": 5,
                    "relay_state": True,
                    "name": "Console",
                    "outlet_caps": 3,
                    "outlet_voltage": "118.566",
                    "outlet_current": "0.061",
                    "outlet_power": "3.815",
                    "outlet_power_factor": "0.527",
                },
                {
                    "index": 6,
                    "relay_state": True,
                    "name": "UDM Pro",
                    "outlet_caps": 3,
                    "outlet_voltage": "118.566",
                    "outlet_current": "0.248",
                    "outlet_power": "14.351",
                    "outlet_power_factor": "0.488",
                },
                {
                    "index": 7,
                    "relay_state": True,
                    "name": "Unraid",
                    "outlet_caps": 3,
                    "outlet_voltage": "118.566",
                    "outlet_current": "1.454",
                    "outlet_power": "169.900",
                    "outlet_power_factor": "0.985",
                },
                {
                    "index": 8,
                    "relay_state": True,
                    "outlet_caps": 3,
                    "outlet_voltage": "118.566",
                    "outlet_current": "0.000",
                    "outlet_power": "0.000",
                    "outlet_power_factor": "0.000",
                    "name": "Outlet 8",
                },
                {
                    "index": 9,
                    "relay_state": True,
                    "outlet_caps": 3,
                    "outlet_voltage": "118.566",
                    "outlet_current": "0.000",
                    "outlet_power": "0.000",
                    "outlet_power_factor": "0.000",
                    "name": "Outlet 9",
                },
                {
                    "index": 10,
                    "relay_state": True,
                    "outlet_caps": 3,
                    "outlet_voltage": "118.566",
                    "outlet_current": "0.000",
                    "outlet_power": "0.000",
                    "outlet_power_factor": "0.000",
                    "name": "Outlet 10",
                },
                {
                    "index": 11,
                    "relay_state": False,
                    "outlet_caps": 3,
                    "outlet_voltage": "118.566",
                    "outlet_current": "0.000",
                    "outlet_power": "0.000",
                    "outlet_power_factor": "0.000",
                    "name": "Outlet 11",
                },
                {
                    "index": 12,
                    "relay_state": True,
                    "outlet_caps": 3,
                    "outlet_voltage": "118.566",
                    "outlet_current": "0.000",
                    "outlet_power": "0.000",
                    "outlet_power_factor": "0.000",
                    "name": "Outlet 12",
                },
                {
                    "index": 13,
                    "relay_state": True,
                    "outlet_caps": 3,
                    "outlet_voltage": "118.566",
                    "outlet_current": "0.000",
                    "outlet_power": "0.000",
                    "outlet_power_factor": "0.000",
                    "name": "Outlet 13",
                },
                {
                    "index": 14,
                    "relay_state": True,
                    "outlet_caps": 3,
                    "outlet_voltage": "118.566",
                    "outlet_current": "0.000",
                    "outlet_power": "0.000",
                    "outlet_power_factor": "0.000",
                    "name": "Outlet 14",
                },
                {
                    "index": 15,
                    "relay_state": True,
                    "outlet_caps": 3,
                    "outlet_voltage": "118.566",
                    "outlet_current": "0.127",
                    "outlet_power": "9.394",
                    "outlet_power_factor": "0.623",
                    "name": "Outlet 15",
                },
                {
                    "index": 16,
                    "relay_state": True,
                    "name": "UNVR Pro",
                    "outlet_caps": 3,
                    "outlet_voltage": "118.566",
                    "outlet_current": "0.278",
                    "outlet_power": "31.992",
                    "outlet_power_factor": "0.970",
                },
                {
                    "index": 17,
                    "relay_state": True,
                    "outlet_caps": 3,
                    "outlet_voltage": "118.566",
                    "outlet_current": "0.000",
                    "outlet_power": "0.000",
                    "outlet_power_factor": "0.000",
                    "name": "Outlet 17",
                },
                {
                    "index": 18,
                    "relay_state": True,
                    "name": "Home Assistant",
                    "outlet_caps": 3,
                    "outlet_voltage": "118.566",
                    "outlet_current": "0.328",
                    "outlet_power": "21.529",
                    "outlet_power_factor": "0.553",
                },
                {
                    "index": 19,
                    "relay_state": True,
                    "name": "Server Cabinet Switch",
                    "outlet_caps": 3,
                    "outlet_voltage": "118.566",
                    "outlet_current": "0.495",
                    "outlet_power": "56.760",
                    "outlet_power_factor": "0.967",
                },
                {
                    "index": 20,
                    "relay_state": True,
                    "name": "Rear Cabinet Lights",
                    "outlet_caps": 3,
                    "outlet_voltage": "118.566",
                    "outlet_current": "0.000",
                    "outlet_power": "0.000",
                    "outlet_power_factor": "0.000",
                },
            ],
            "outlet_overrides": [
                {
                    "index": 1,
                    "name": "USB Outlet 1",
                    "cycle_enabled": False,
                    "relay_state": False,
                },
                {
                    "index": 2,
                    "name": "USB Outlet 2",
                    "cycle_enabled": False,
                    "relay_state": False,
                },
                {
                    "index": 3,
                    "name": "USB Outlet 3",
                    "cycle_enabled": False,
                    "relay_state": False,
                },
                {
                    "index": 4,
                    "name": "USB Outlet 4",
                    "cycle_enabled": False,
                    "relay_state": False,
                },
                {"index": 5, "name": "Console", "relay_state": True},
                {"index": 6, "name": "UDM Pro", "relay_state": True},
                {"index": 7, "name": "Unraid", "relay_state": True},
                {"index": 8, "relay_state": True},
                {"index": 9, "relay_state": True},
                {"index": 10, "relay_state": True},
                {"index": 11, "relay_state": True},
                {"index": 12, "relay_state": True},
                {"index": 13, "relay_state": True},
                {"index": 14, "relay_state": True},
                {"index": 15, "relay_state": True},
                {"index": 16, "name": "UNVR Pro", "relay_state": True},
                {"index": 17, "relay_state": True},
                {"index": 18, "name": "Home Assistant", "relay_state": True},
                {"index": 19, "name": "Server Cabinet Switch", "relay_state": True},
                {"index": 20, "name": "Rear Cabinet Lights", "relay_state": True},
            ],
            "port_table": [
                {
                    "port_idx": 1,
                    "media": "FE",
                    "port_poe": False,
                    "poe_caps": 0,
                    "speed_caps": 1048591,
                    "op_mode": "switch",
                    "portconf_id": "5fc7fb23c3da2e039ebeea97",
                    "autoneg": False,
                    "enable": True,
                    "flowctrl_rx": False,
                    "flowctrl_tx": False,
                    "full_duplex": True,
                    "is_uplink": True,
                    "jumbo": False,
                    "mac_table": [],
                    "rx_broadcast": 0,
                    "rx_bytes": 538000102,
                    "rx_dropped": 2,
                    "rx_errors": 0,
                    "rx_multicast": 0,
                    "rx_packets": 3943979,
                    "satisfaction": 90,
                    "satisfaction_reason": 1,
                    "speed": 100,
                    "stp_pathcost": 0,
                    "stp_state": "disabled",
                    "tx_broadcast": 0,
                    "tx_bytes": 114523726,
                    "tx_dropped": 0,
                    "tx_errors": 0,
                    "tx_multicast": 0,
                    "tx_packets": 670312,
                    "up": True,
                    "tx_bytes-r": 82,
                    "rx_bytes-r": 1510,
                    "bytes-r": 1592,
                    "name": "Port 1",
                    "masked": False,
                    "aggregated_by": False,
                }
            ],
            "state": 1,
            "sys_stats": {
                "loadavg_1": "0.08",
                "loadavg_15": "0.01",
                "loadavg_5": "0.02",
                "mem_buffer": 0,
                "mem_total": 61792256,
                "mem_used": 18235392,
            },
            "system_stats": ("1.1", "29.5", "1347471"),
            "type": "usw",
            "version": "5.76.7.13442",
            "upgradable": False,
            "uplink": PDU_PRO["uplink"],
            "uptime": 1347471,
        },
    ),
    (
        [SWITCH_16_PORT_POE],
        {
            "board_revision": 9,
            "considered_lost_at": 1588175821,
            "disabled": False,
            "general_temperature": 56,
            "has_temperature": True,
            "id": "235678987654345678",
            "ip": "192.168.0.57",
            "downlink_table": [],
            "fan_level": 0,
            "has_fan": True,
            "last_seen": 1588175722,
            "lldp_table": [],
            "mac": "fc:ec:da:11:22:33",
            "model": "US16P150",
            "name": "Switch 16",
            "next_interval": 23,
            "next_heartbeat_at": 1588175755,
            "overheating": False,
            "port_overrides": [
                {
                    "poe_mode": "auto",
                    "portconf_id": "5e1b309714bd614afd3d11a7",
                    "port_security_mac_address": [],
                    "autoneg": True,
                    "stp_port_mode": True,
                },
                {
                    "poe_mode": "off",
                    "port_idx": 3,
                    "portconf_id": "5a32aa4ee4babd4452422ddd22222",
                },
                {
                    "poe_mode": "auto",
                    "port_idx": 4,
                    "portconf_id": "5a32aa4ee4babd4452422ddd22222",
                },
                {
                    "poe_mode": "auto",
                    "port_idx": 16,
                    "portconf_id": "5a32aa4ee4babd4452422ddd22222",
                },
            ],
            "port_table": SWITCH_16_PORT_POE["port_table"],
            "state": 1,
            "sys_stats": {
                "loadavg_1": "2.82",
                "loadavg_15": "2.81",
                "loadavg_5": "2.80",
                "mem_buffer": 0,
                "mem_total": 262402048,
                "mem_used": 129331200,
            },
            "system_stats": ("100.0", "49.3", "3971865"),
            "type": "usw",
            "version": "4.0.66.10832",
            "upgradable": True,
            "upgrade_to_firmware": "4.0.80.10875",
            "uplink": SWITCH_16_PORT_POE["uplink"],
            "uplink_depth": 2,
            "uptime": 3971865,
            "user_num_sta": 4,
            "wlan_overrides": [],
        },
    ),
]


@pytest.mark.parametrize(("device_payload", "reference_data"), test_data)
async def test_device(
    unifi_controller: Controller, _mock_endpoints: None, reference_data: dict[str, Any]
) -> None:
    """Test device class."""
    devices = unifi_controller.devices
    await devices.update()
    assert len(devices.items()) == 1

    device = next(iter(devices.values()))
    for key, value in reference_data.items():
        assert getattr(device, key) == value
    assert repr(device) == f"<Device {device.name}: {device.mac}>"


@pytest.mark.parametrize(
    ("method", "mac", "command"),
    [["upgrade", "0", {"mac": "0", "cmd": "upgrade"}]],
)
async def test_device_commands(
    mock_aioresponse: aioresponses,
    unifi_controller: Controller,
    unifi_called_with: Callable[[str, str, dict[str, Any]], bool],
    method: str,
    mac: str,
    command: dict[str, Any],
) -> None:
    """Test device commands."""
    mock_aioresponse.post("https://host:8443/api/s/default/cmd/devmgr", payload={})
    class_command = getattr(unifi_controller.devices, method)
    await class_command(mac)
    assert unifi_called_with("post", "/api/s/default/cmd/devmgr", json=command)


@pytest.mark.parametrize(
    ("api_request", "data", "command"),
    [
        [
            DeviceRestartRequest,
            {"mac": "0", "soft": True},
            {"mac": "0", "cmd": "restart", "reboot_type": "soft"},
        ],
        [
            DeviceRestartRequest,
            {"mac": "0", "soft": False},
            {"mac": "0", "cmd": "restart", "reboot_type": "hard"},
        ],
        [
            DeviceUpgradeRequest,
            {"mac": "0"},
            {"mac": "0", "cmd": "upgrade"},
        ],
        [
            DevicePowerCyclePortRequest,
            {"mac": "0", "port_idx": 1},
            {"mac": "0", "port_idx": 1, "cmd": "power-cycle"},
        ],
    ],
)
async def test_device_requests(
    mock_aioresponse: aioresponses,
    unifi_controller: Controller,
    unifi_called_with: Callable[[str, str, dict[str, Any]], bool],
    api_request: DeviceRestartRequest
    | DeviceUpgradeRequest
    | DevicePowerCyclePortRequest,
    data: dict[str, Any],
    command: dict[str, Any],
) -> None:
    """Test device commands."""
    mock_aioresponse.post("https://host:8443/api/s/default/cmd/devmgr", payload={})
    await unifi_controller.request(api_request.create(**data))
    assert unifi_called_with("post", "/api/s/default/cmd/devmgr", json=command)


@pytest.mark.parametrize(
    ("device_payload", "api_request", "data", "command"),
    [
        [  # Outlet set relay without existing override
            [
                {
                    "device_id": "01",
                    "mac": "0",
                    "outlet_overrides": [],
                    "outlet_table": [
                        {
                            "index": 1,
                            "relay_state": True,
                            "cycle_enabled": False,
                            "name": "USB Outlet 1",
                            "outlet_caps": 1,
                        },
                    ],
                }
            ],
            DeviceSetOutletRelayRequest,
            {"outlet_idx": 1, "state": True},
            {
                "outlet_overrides": [
                    {"index": 1, "name": "USB Outlet 1", "relay_state": True}
                ]
            },
        ],
        [  # Outlet set relay with existing override
            [
                {
                    "device_id": "01",
                    "mac": "0",
                    "outlet_overrides": [{"index": 2}],
                    "outlet_table": [
                        {
                            "index": 2,
                            "relay_state": True,
                            "cycle_enabled": False,
                            "name": "USB Outlet 1",
                            "outlet_caps": 1,
                        },
                    ],
                }
            ],
            DeviceSetOutletRelayRequest,
            {"outlet_idx": 2, "state": False},
            {"outlet_overrides": [{"index": 2, "relay_state": False}]},
        ],
        [  # Outlet outlet cycle without existing override
            [
                {
                    "device_id": "01",
                    "mac": "0",
                    "outlet_overrides": [],
                    "outlet_table": [
                        {
                            "index": 1,
                            "relay_state": True,
                            "cycle_enabled": False,
                            "name": "USB Outlet 1",
                            "outlet_caps": 1,
                        },
                    ],
                }
            ],
            DeviceSetOutletCycleEnabledRequest,
            {"outlet_idx": 1, "state": True},
            {
                "outlet_overrides": [
                    {"index": 1, "name": "USB Outlet 1", "cycle_enabled": True}
                ]
            },
        ],
        [  # Outlet outlet cycle with existing override
            [
                {
                    "device_id": "01",
                    "mac": "0",
                    "outlet_overrides": [{"index": 2}],
                    "outlet_table": [
                        {
                            "index": 2,
                            "relay_state": True,
                            "cycle_enabled": False,
                            "name": "USB Outlet 1",
                            "outlet_caps": 1,
                        },
                    ],
                }
            ],
            DeviceSetOutletCycleEnabledRequest,
            {"outlet_idx": 2, "state": False},
            {"outlet_overrides": [{"index": 2, "cycle_enabled": False}]},
        ],
        [  # PoE port mode without existing override
            [
                {
                    "device_id": "01",
                    "mac": "0",
                    "port_overrides": [],
                    "port_table": [
                        {
                            "poe_mode": "Auto",
                            "name": "Port 1",
                            "port_idx": 1,
                        },
                    ],
                }
            ],
            DeviceSetPoePortModeRequest,
            {"port_idx": 1, "mode": "off"},
            {"port_overrides": [{"port_idx": 1, "poe_mode": "off"}]},
        ],
        [  # PoE port mode with portconf_id without existing override
            [
                {
                    "device_id": "01",
                    "mac": "0",
                    "port_overrides": [],
                    "port_table": [
                        {
                            "poe_mode": "Auto",
                            "name": "Port 1",
                            "port_idx": 1,
                            "portconf_id": "123",
                        },
                    ],
                }
            ],
            DeviceSetPoePortModeRequest,
            {"port_idx": 1, "mode": "off"},
            {
                "port_overrides": [
                    {"port_idx": 1, "poe_mode": "off", "portconf_id": "123"}
                ]
            },
        ],
        [  # PoE port mode with existing override
            [
                {
                    "device_id": "01",
                    "mac": "0",
                    "port_overrides": [{"port_idx": 1, "name": "Office"}],
                    "port_table": [
                        {
                            "poe_mode": "Auto",
                            "name": "Office",
                            "port_idx": 1,
                        },
                    ],
                }
            ],
            DeviceSetPoePortModeRequest,
            {"port_idx": 1, "mode": "off"},
            {"port_overrides": [{"port_idx": 1, "poe_mode": "off", "name": "Office"}]},
        ],
        [  # PoE multi target port mode with existing override
            [
                {
                    "device_id": "01",
                    "mac": "0",
                    "port_overrides": [{"port_idx": 1, "name": "Office"}],
                    "port_table": [
                        {
                            "poe_mode": "Auto",
                            "name": "Office",
                            "port_idx": 1,
                        },
                        {
                            "poe_mode": "Off",
                            "name": "Hallway",
                            "port_idx": 2,
                        },
                    ],
                }
            ],
            DeviceSetPoePortModeRequest,
            {"targets": [(1, "off"), (2, "auto")]},
            {
                "port_overrides": [
                    {"port_idx": 1, "poe_mode": "off", "name": "Office"},
                    {"port_idx": 2, "poe_mode": "auto"},
                ]
            },
        ],
    ],
)
async def test_sub_device_requests(
    mock_aioresponse: aioresponses,
    unifi_controller: Controller,
    _mock_endpoints: None,
    unifi_called_with: Callable[[str, str, dict[str, Any]], bool],
    api_request: DeviceSetOutletRelayRequest
    | DeviceSetOutletCycleEnabledRequest
    | DeviceSetPoePortModeRequest,
    data: dict[str, Any],
    command: dict[str, Any],
) -> None:
    """Test sub device (port/outlet) commands."""
    devices = unifi_controller.devices
    await devices.update()
    device = next(iter(devices.values()))
    mock_aioresponse.put("https://host:8443/api/s/default/rest/device/01", payload={})
    await unifi_controller.request(api_request.create(device, **data))
    assert unifi_called_with("put", "/api/s/default/rest/device/01", json=command)


@pytest.mark.parametrize(("device_payload"), [[SWITCH_16_PORT_POE]])
async def test_set_poe_request_raise_error(
    unifi_controller: Controller, _mock_endpoints: None
) -> None:
    """Test device class."""
    await unifi_controller.initialize()
    device = next(iter(unifi_controller.devices.values()))
    with pytest.raises(AttributeError):
        DeviceSetPoePortModeRequest.create(device)


async def test_device_websocket(
    unifi_controller: Controller, _new_ws_data_fn: Callable[[dict[str, Any]], None]
) -> None:
    """Test controller managing devices."""
    assert len(unifi_controller.devices._subscribers["*"]) == 2

    unsub = unifi_controller.devices.subscribe(mock_callback := Mock())
    assert len(unifi_controller.devices._subscribers["*"]) == 3
    assert mock_callback.call_count == 0

    # Add client from websocket
    _new_ws_data_fn(
        {
            "meta": {"message": MessageKey.DEVICE.value},
            "data": [SWITCH_16_PORT_POE],
        }
    )
    assert len(unifi_controller.devices.items()) == 1
    assert len(unifi_controller.devices._subscribers["*"]) == 3

    unsub()
    assert len(unifi_controller.devices._subscribers["*"]) == 2


def test_enum_unknowns() -> None:
    """Validate enum unknown values."""
    assert DeviceState(999) == DeviceState.UNKNOWN



"""Test aioUniFi device outlets.

pytest --cov-report term-missing --cov=aiounifi.outlets tests/test_outlets.py
"""

from unittest.mock import Mock

import pytest

from aiounifi.interfaces.api_handlers import ItemEvent
from aiounifi.models.outlet import Outlet

from .fixtures import PDU_PRO, PLUG_UP1, STRIP_UP6


async def test_handler_ports(unifi_controller):
    """Verify that device ports works."""
    outlets = unifi_controller.outlets
    assert len(outlets.items()) == 0
    assert len(outlets.values()) == 0

    assert "78:45:58:fc:16:7d_1" not in outlets
    assert outlets.get("78:45:58:fc:16:7d_1", "xy") == "xy"
    assert not isinstance(outlets.get("78:45:58:fc:16:7d_1"), Outlet)
    unsub_all = outlets.subscribe(mock_subscribe_all := Mock())
    unsub_7 = outlets.subscribe(
        mock_subscribe_7 := Mock(), id_filter="78:45:58:fc:16:7d_7"
    )
    unsub_added = outlets.subscribe(
        mock_subscribe_added := Mock(), event_filter=ItemEvent.ADDED
    )
    unsub_changed = outlets.subscribe(
        mock_subscribe_changed := Mock(), event_filter=ItemEvent.CHANGED
    )
    unsub_deleted = outlets.subscribe(
        mock_subscribe_deleted := Mock(), event_filter=ItemEvent.DELETED
    )
    unsub_bad = outlets.subscribe(mock_subscribe_bad := Mock(), id_filter="bad")

    # Add outlets
    unifi_controller.devices.process_raw([STRIP_UP6])
    assert next(iter(outlets)) == "78:45:58:fc:16:7d_1"
    assert "78:45:58:fc:16:7d_2" in outlets
    assert isinstance(outlets.get("78:45:58:fc:16:7d_3"), Outlet)
    assert isinstance(outlets["78:45:58:fc:16:7d_4"], Outlet)
    assert len(outlets.values()) == 7
    assert mock_subscribe_all.call_count == 7
    mock_subscribe_all.assert_called_with(ItemEvent.ADDED, "78:45:58:fc:16:7d_7")
    mock_subscribe_7.assert_called_once()
    assert mock_subscribe_added.call_count == 7
    mock_subscribe_changed.assert_not_called()
    mock_subscribe_deleted.assert_not_called()
    mock_subscribe_bad.assert_not_called()

    # Update outlets
    unifi_controller.devices.process_raw([STRIP_UP6])
    assert len(outlets.values()) == 7
    assert mock_subscribe_all.call_count == 14
    mock_subscribe_all.assert_called_with(ItemEvent.CHANGED, "78:45:58:fc:16:7d_7")
    assert mock_subscribe_7.call_count == 2
    assert mock_subscribe_added.call_count == 7
    assert mock_subscribe_changed.call_count == 7
    mock_subscribe_deleted.assert_not_called()
    mock_subscribe_bad.assert_not_called()

    # Remove outlets
    unifi_controller.devices.remove_item(STRIP_UP6)
    assert len(outlets.values()) == 0
    assert mock_subscribe_all.call_count == 21
    mock_subscribe_all.assert_called_with(ItemEvent.DELETED, "78:45:58:fc:16:7d_7")
    assert mock_subscribe_7.call_count == 3
    assert mock_subscribe_added.call_count == 7
    assert mock_subscribe_changed.call_count == 7
    assert mock_subscribe_deleted.call_count == 7
    mock_subscribe_bad.assert_not_called()

    unsub_all()
    unsub_7()
    unsub_added()
    unsub_changed()
    unsub_deleted()
    unsub_bad()

    unsub_all()
    unsub_all()


async def test_handler_process_device_no_index(unifi_controller):
    """Verify that device ports works."""
    ports = unifi_controller.ports
    unifi_controller.devices.process_raw([{"mac": "1", "outlet_table": []}])
    assert len(ports.items()) == 0


data_test_outlet = [
    (
        [PDU_PRO],
        {
            "device": PDU_PRO,
            "obj_id": "00:00:00:00:00:84_1",
            "name": "USB Outlet 1",
            "index": 1,
            "relay_state": True,
            "has_relay": None,
            "cycle_enabled": False,
            "has_metering": None,
            "caps": 1,
            "voltage": None,
            "current": None,
            "power": None,
            "power_factor": None,
            "repr": "<USB Outlet 1: relay state True>",
        },
    ),
    (
        [PDU_PRO],
        {
            "device": PDU_PRO,
            "obj_id": "00:00:00:00:00:84_5",
            "name": "Console",
            "index": 5,
            "relay_state": True,
            "has_relay": None,
            "cycle_enabled": None,
            "has_metering": None,
            "caps": 3,
            "voltage": "118.566",
            "current": "0.061",
            "power": "3.815",
            "power_factor": "0.527",
            "repr": "<Console: relay state True>",
        },
    ),
    (
        [PLUG_UP1],
        {
            "device": PLUG_UP1,
            "obj_id": "fc:ec:da:76:4f:5f_1",
            "name": "Outlet 1",
            "index": 1,
            "relay_state": False,
            "has_relay": True,
            "cycle_enabled": None,
            "has_metering": False,
            "caps": None,
            "voltage": None,
            "current": None,
            "power": None,
            "power_factor": None,
            "repr": "<Outlet 1: relay state False>",
        },
    ),
    (
        [STRIP_UP6],
        {
            "device": STRIP_UP6,
            "obj_id": "78:45:58:fc:16:7d_1",
            "name": "Outlet 1",
            "index": 1,
            "relay_state": False,
            "has_relay": True,
            "cycle_enabled": False,
            "has_metering": False,
            "caps": None,
            "voltage": None,
            "current": None,
            "power": None,
            "power_factor": None,
            "repr": "<Outlet 1: relay state False>",
        },
    ),
]


@pytest.mark.parametrize(("device_payload", "test_data"), data_test_outlet)
async def test_outlet(unifi_controller, _mock_endpoints, test_data):
    """Verify that device outlet model works."""
    await unifi_controller.devices.update()
    outlet = unifi_controller.outlets[test_data["obj_id"]]

    assert outlet.name == test_data["name"]
    assert outlet.index == test_data["index"]
    assert outlet.relay_state == test_data["relay_state"]
    assert outlet.has_relay == test_data["has_relay"]
    assert outlet.cycle_enabled == test_data["cycle_enabled"]
    assert outlet.has_metering == test_data["has_metering"]
    assert outlet.caps == test_data["caps"]
    assert outlet.voltage == test_data["voltage"]
    assert outlet.current == test_data["current"]
    assert outlet.power == test_data["power"]
    assert outlet.power_factor == test_data["power_factor"]

    assert repr(outlet) == test_data["repr"]



"""UniFi sites of network infrastructure."""

from ..models.site import Site, SiteListRequest
from .api_handlers import APIHandler


class Sites(APIHandler[Site]):
    """Represent UniFi sites."""

    obj_id_key = "_id"
    item_cls = Site
    api_request = SiteListRequest.create()



"""Device port implementation."""


from .api import ApiItem
from .device import TypedDevicePortTable


class Port(ApiItem):
    """Represents a network port."""

    raw: TypedDevicePortTable

    @property
    def ifname(self) -> str | None:
        """Port name used by USG."""
        return self.raw.get("ifname")

    @property
    def media(self) -> str | None:
        """Media port is connected to."""
        return self.raw.get("media")

    @property
    def name(self) -> str:
        """Port name."""
        return self.raw["name"]

    @property
    def port_idx(self) -> int | None:
        """Port index."""
        return self.raw.get("port_idx")

    @property
    def poe_caps(self) -> int | None:
        """Port PoE capabilities.

        0 - no caps
        3 - auto (PoE/PoE+)
        35 - auto (Poe/PoE+/PoE++)
        7 - 24V passive
        8 - Passthrough
        """
        return self.raw.get("poe_caps")

    @property
    def poe_class(self) -> str | None:
        """Port PoE class."""
        return self.raw.get("poe_class")

    @property
    def poe_enable(self) -> bool | None:
        """Is PoE supported/requested by client."""
        return self.raw.get("poe_enable")

    @property
    def poe_mode(self) -> str | None:
        """Is PoE auto, pasv24, passthrough, off or None."""
        return self.raw.get("poe_mode")

    @property
    def poe_power(self) -> str | None:
        """Is PoE power usage."""
        return self.raw.get("poe_power")

    @property
    def poe_voltage(self) -> str | None:
        """Is PoE voltage usage."""
        return self.raw.get("poe_voltage")

    @property
    def portconf_id(self) -> str | None:
        """Port configuration ID."""
        return self.raw.get("portconf_id")

    @property
    def port_poe(self) -> bool | None:
        """Is PoE used."""
        return self.raw.get("port_poe")

    @property
    def up(self) -> bool | None:
        """Is port up."""
        return self.raw.get("up")

    def __repr__(self) -> str:
        """Return the representation."""
        return f"<{self.name}: Poe {self.poe_enable}>"



"""Test port forwarding API."""
import pytest

from aiounifi.controller import Controller
from aiounifi.models.port_forward import PortForwardEnableRequest, TypedPortForward

from .fixtures import PORT_FORWARDING


@pytest.mark.parametrize("port_forward_payload", [PORT_FORWARDING["data"]])
async def test_port_forward(
    mock_aioresponse, unifi_controller: Controller, _mock_endpoints, unifi_called_with
):
    """Test port forwarding interface and model."""
    port_forwarding = unifi_controller.port_forwarding
    await port_forwarding.update()
    assert len(port_forwarding.values()) == 1

    port_forward = next(iter(port_forwarding.values()))
    assert port_forward.id == "5a32aa4ee4b0412345678911"
    assert port_forward.destination_port == "12345"
    assert port_forward.enabled is True
    assert port_forward.forward_port == "23456"
    assert port_forward.forward_ip == "10.0.0.2"
    assert port_forward.name == "New port forward"
    assert port_forward.port_forward_interface == "wan"
    assert port_forward.protocol == "tcp_udp"
    assert port_forward.site_id == "5a32aa4ee4b0412345678910"
    assert port_forward.source == "any"

    # Port forward enable request

    pf_id = PORT_FORWARDING["data"][0]["_id"]
    mock_aioresponse.put(
        f"https://host:8443/api/s/default/rest/portforward/{pf_id}",
        payload={},
    )

    expected: TypedPortForward = PORT_FORWARDING["data"][0].copy()
    expected["enabled"] = False

    await unifi_controller.request(PortForwardEnableRequest.create(port_forward, False))

    assert unifi_called_with(
        "put",
        f"/api/s/default/rest/portforward/{pf_id}",
        json=expected,
    )
    assert port_forward.raw != expected



"""System information of a UniFi network."""

from ..models.system_information import SystemInformation, SystemInformationRequest
from .api_handlers import APIHandler


class SystemInformationHandler(APIHandler[SystemInformation]):
    """Represents system information interface."""

    obj_id_key = "anonymous_controller_id"
    item_cls = SystemInformation
    api_request = SystemInformationRequest.create()



"""Python library to enable integration between Home Assistant and UniFi."""

from dataclasses import KW_ONLY, dataclass
from ssl import SSLContext
from typing import Literal

from aiohttp import ClientSession


@dataclass
class Configuration:
    """Console configuration."""

    session: ClientSession
    host: str
    _: KW_ONLY
    username: str
    password: str
    port: int = 8443
    site: str = "default"
    ssl_context: SSLContext | Literal[False] = False

    @property
    def url(self) -> str:
        """Represent console path."""
        return f"https://{self.host}:{self.port}"



"""Clients are devices on a UniFi network."""

from ..models.client import AllClientListRequest, Client
from .api_handlers import APIHandler


class ClientsAll(APIHandler[Client]):
    """Represents all client network devices."""

    obj_id_key = "mac"
    item_cls = Client
    api_request = AllClientListRequest.create()



"""Site is a specific grouping in a UniFi network."""
from dataclasses import dataclass
from typing import Self, TypedDict

from .api import ApiItem, ApiRequest


class TypedSite(TypedDict):
    """Site description."""

    _id: str
    attr_hidden_id: str
    attr_no_delete: bool
    desc: str
    name: str
    role: str


@dataclass
class SiteListRequest(ApiRequest):
    """Request object for site list."""

    @classmethod
    def create(cls) -> Self:
        """Create site list request."""
        return cls(method="get", path="/self/sites")

    def full_path(self, site: str, is_unifi_os: bool) -> str:
        """Url to list sites is global for controller."""
        if is_unifi_os:
            return f"/proxy/network/api{self.path}"
        return f"/api{self.path}"


class Site(ApiItem):
    """Represents a network device."""

    raw: TypedSite

    @property
    def site_id(self) -> str:
        """Site ID."""
        return self.raw["_id"]

    @property
    def description(self) -> str:
        """Site description."""
        return self.raw["desc"]

    @property
    def hidden_id(self) -> str:
        """Unknown."""
        return self.raw["attr_hidden_id"]

    @property
    def name(self) -> str:
        """Site name."""
        return self.raw["name"]

    @property
    def no_delete(self) -> bool:
        """Can not delete site."""
        return self.raw["attr_no_delete"]

    @property
    def role(self) -> str:
        """User role."""
        return self.raw["role"]



"""API management class and base class for the different end points."""

from abc import ABC
from collections.abc import Mapping
from dataclasses import dataclass
from typing import Any, TypedDict, TypeVar

import orjson

from ..errors import (
    AiounifiException,
    LoginRequired,
    NoPermission,
    TwoFaTokenRequired,
    Unauthorized,
)

ERRORS = {
    "api.err.Invalid": Unauthorized,
    "api.err.LoginRequired": LoginRequired,
    "api.err.NoPermission": NoPermission,
    "api.err.Ubic2faTokenRequired": TwoFaTokenRequired,
}


class TypedApiResponse(TypedDict, total=False):
    """Common response."""

    meta: dict[str, Any]
    data: list[dict[str, Any]]


@dataclass
class ApiRequest:
    """Data class with required properties of a request."""

    method: str
    path: str
    data: Mapping[str, Any] | None = None

    def full_path(self, site: str, is_unifi_os: bool) -> str:
        """Create url to work with a specific controller."""
        if is_unifi_os:
            return f"/proxy/network/api/s/{site}{self.path}"
        return f"/api/s/{site}{self.path}"

    def decode(self, raw: bytes) -> TypedApiResponse:
        """Put data, received from the unifi controller, into a TypedApiResponse."""
        data: TypedApiResponse = orjson.loads(raw)

        if "meta" in data and data["meta"]["rc"] == "error":
            raise ERRORS.get(data["meta"]["msg"], AiounifiException)(data)

        return data


@dataclass
class ApiRequestV2(ApiRequest):
    """Data class with required properties of a V2 API request."""

    def full_path(self, site: str, is_unifi_os: bool) -> str:
        """Create url to work with a specific controller."""
        if is_unifi_os:
            return f"/proxy/network/v2/api/site/{site}{self.path}"
        return f"/v2/api/site/{site}{self.path}"

    def decode(self, raw: bytes) -> TypedApiResponse:
        """Put data, received from the unifi controller, into a TypedApiResponse."""
        data = orjson.loads(raw)

        if "errorCode" in data:
            raise ERRORS.get(data["message"], AiounifiException)(data)

        return TypedApiResponse(
            meta={"rc": "ok", "msg": ""},
            data=data if isinstance(data, list) else [data],
        )


# @dataclass
# class ApiItem(ABC):
#     """Base class for all end points using APIItems class."""

#     raw: Any


class ApiItem(ABC):
    """Base class for all end points using APIItems class."""

    def __init__(self, raw: Any) -> None:
        """Initialize API item."""
        self.raw = raw


ApiItemT = TypeVar("ApiItemT", bound=ApiItem)



"""Traffic rules as part of a UniFi network."""

from dataclasses import dataclass
from typing import NotRequired, Self, TypedDict

from .api import ApiItem, ApiRequestV2


class BandwidthLimit(TypedDict):
    """Bandwidth limit type definition."""

    download_limit_kbps: int
    enabled: bool
    upload_limit_kbps: int


class PortRange(TypedDict):
    """Port range type definition."""

    port_start: int
    port_stop: int


class IPAddress(TypedDict):
    """IP Address for which traffic rule is applicable type definition."""

    ip_or_subnet: str
    ip_version: str
    port_ranges: list[PortRange]
    ports: list[int]


class IPRange(TypedDict):
    """IP Range type definition."""

    ip_start: str
    ip_stop: str
    ip_version: str


class Schedule(TypedDict):
    """Schedule to enable/disable traffic rule type definition."""

    date_end: str
    date_start: str
    mode: str
    repeat_on_days: list[str]
    time_all_day: bool
    time_range_end: str
    time_range_start: str


class TargetDevice(TypedDict):
    """Target device to which the traffic rule applies."""

    client_mac: NotRequired[str]
    network_id: NotRequired[str]
    type: str


class TypedTrafficRule(TypedDict):
    """Traffic rule type definition."""

    _id: str
    action: str
    app_category_ids: list[str]
    app_ids: list[str]
    bandwidth_limit: BandwidthLimit
    description: str
    domains: list[str]
    enabled: bool
    ip_addresses: list[IPAddress]
    ip_ranges: list[IPRange]
    matching_target: str
    network_ids: list[str]
    regions: list[str]
    schedule: Schedule
    target_devices: list[TargetDevice]


@dataclass
class TrafficRuleListRequest(ApiRequestV2):
    """Request object for traffic rule list."""

    @classmethod
    def create(cls) -> Self:
        """Create traffic rule request."""
        return cls(method="get", path="/trafficrules", data=None)


@dataclass
class TrafficRuleEnableRequest(ApiRequestV2):
    """Request object for traffic rule enable."""

    @classmethod
    def create(cls, traffic_rule: TypedTrafficRule, enable: bool) -> Self:
        """Create traffic rule enable request."""
        traffic_rule["enabled"] = enable
        return cls(
            method="put",
            path=f"/trafficrules/{traffic_rule['_id']}",
            data=traffic_rule,
        )


class TrafficRule(ApiItem):
    """Represent a traffic rule configuration."""

    raw: TypedTrafficRule

    @property
    def id(self) -> str:
        """ID of traffic rule."""
        return self.raw["_id"]

    @property
    def description(self) -> str:
        """Description given by user to traffic rule."""
        return self.raw["description"]

    @property
    def enabled(self) -> bool:
        """Is traffic rule enabled."""
        return self.raw["enabled"]

    @property
    def action(self) -> str:
        """What action is defined by this traffic rule."""
        return self.raw["action"]

    @property
    def matching_target(self) -> str:
        """What target is matched by this traffic rule."""
        return self.raw["matching_target"]

    @property
    def target_devices(self) -> list[TargetDevice]:
        """What target devices are affected by this traffic rule."""
        return self.raw["target_devices"]



"""Test events.

pytest --cov-report term-missing --cov=aiounifi.events tests/test_events.py
"""

from unittest.mock import Mock

import pytest

from aiounifi.interfaces.events import EventHandler
from aiounifi.models.event import Event, EventKey
from aiounifi.models.message import Message, MessageKey, Meta

from .fixtures import EVENT_SWITCH_16_CONNECTED, EVENT_WIRELESS_CLIENT_CONNECTED

EVENT_HANDLER_DATA = [
    (None, True),  # No filters
    (EventKey.SWITCH_LOST_CONTACT, True),  # Filter correct
    (EventKey.SWITCH_CONNECTED, False),  # Filter incorrect
]


@pytest.mark.parametrize(("event_filter", "expected"), EVENT_HANDLER_DATA)
async def test_event_handler(event_filter, expected):
    """Verify event handler behaves according to configured filters."""
    event_handler = EventHandler(controller=Mock())

    filters = {}
    if event_filter:
        filters["event_filter"] = event_filter

    unsubscribe_callback = event_handler.subscribe(mock_callback := Mock(), **filters)
    assert len(event_handler) == 1
    assert unsubscribe_callback

    event_handler.handler(
        Message(
            meta=Meta("ok", MessageKey.EVENT, {}),
            data={
                "_id": "5eae7fe02ab79c00f9d38960",
                "datetime": "2020-05-03T08:25:04Z",
                "key": "EVT_SW_Lost_Contact",
                "msg": "Switch[00:..:00] was disconnected",
                "site_id": "default",
                "subsystem": "lan",
                "sw": "00:..:00",
                "sw_name": "switch name",
                "time": 1588494304030,
            },
        ),
    )
    assert mock_callback.called is expected

    unsubscribe_callback()
    assert len(event_handler) == 0


async def test_unsupported_event_key():
    """Test empty event."""
    event = Event({"key": "unsupported"})
    assert event.key == EventKey.UNKNOWN


async def test_empty_event():
    """Test empty event."""
    empty = Event({})

    with pytest.raises(KeyError):
        assert empty.event
    with pytest.raises(KeyError):
        assert empty.key
    with pytest.raises(KeyError):
        assert empty.datetime
    with pytest.raises(KeyError):
        assert empty.msg
    with pytest.raises(KeyError):
        assert empty.time
    assert empty.mac == ""
    assert empty.ap == ""
    assert empty.bytes == 0
    assert empty.channel == 0
    assert empty.duration == 0
    assert empty.hostname == ""
    assert empty.radio == ""
    assert empty.subsystem == ""
    assert empty.site_id == ""
    assert empty.ssid == ""


async def test_client_event():
    """Test client event."""
    event_data = EVENT_WIRELESS_CLIENT_CONNECTED["data"][0]
    client = Event(event_data)

    assert client.event == "EVT_WU_Connected"
    assert client.key == EventKey.WIRELESS_CLIENT_CONNECTED
    assert client.datetime == "2020-04-24T18:37:36Z"
    assert (
        client.msg
        == "User[00:00:00:00:00:01] has connected to AP[80:2a:a8:00:01:02] "
        + 'with SSID "SSID" on "channel 44(na)"'
    )
    assert client.time == 1587753456179
    assert client.mac == "00:00:00:00:00:01"
    assert client.ap == "80:2a:a8:00:01:02"
    assert client.bytes == 0
    assert client.channel == 44
    assert client.duration == 0
    assert client.hostname == "client"
    assert client.radio == "na"
    assert client.subsystem == "wlan"
    assert client.site_id == "5a32aa4ee4b0412345678910"
    assert client.ssid == "SSID"


async def test_device_event():
    """Test device event."""
    event_data = EVENT_SWITCH_16_CONNECTED["data"][0]
    device = Event(event_data)

    assert device.event == "EVT_SW_Connected"
    assert device.key == EventKey.SWITCH_CONNECTED
    assert device.datetime == "2020-05-03T08:35:35Z"
    assert device.msg == "Switch[fc:ec:da:11:22:33] was connected"
    assert device.time == 1588494935241
    assert device.mac == "fc:ec:da:11:22:33"
    assert device.ap == ""
    assert device.bytes == 0
    assert device.channel == 0
    assert device.duration == 0
    assert device.hostname == ""
    assert device.radio == ""
    assert device.subsystem == "lan"
    assert device.site_id == "5a32aa4ee4b0412345678910"
    assert device.ssid == ""
    assert device.version_from == ""
    assert device.version_to == ""



"""Messages from websocket."""

#  https://demo.ui.com/manage/locales/en/eventStrings.json?v=5.4.11.2

from dataclasses import dataclass
import enum
import logging
from typing import Any, Self

LOGGER = logging.getLogger(__name__)


class MessageKey(enum.Enum):
    """Message as part of meta object of event.

    "meta": {"rc": "ok", "message": "device:sync"}.
    """

    ALERT = "alert"
    EVENT = "events"

    CLIENT = "sta:sync"
    CLIENT_REMOVED = "user:delete"
    CLIENT_UPDATED = "user:sync"
    DEVICE = "device:sync"
    DEVICE_ADD = "device:add"
    DEVICE_UPDATE = "device:update"
    DPI_APP_ADDED = "dpiapp:add"
    DPI_APP_REMOVED = "dpiapp:delete"
    DPI_APP_UPDATED = "dpiapp:sync"
    DPI_GROUP_ADDED = "dpigroup:add"
    DPI_GROUP_REMOVED = "dpigroup:delete"
    DPI_GROUP_UPDATED = "dpigroup:sync"
    FIREWALL_RULE_ADDED = "firewallrule:add"
    FIREWALL_RULE_UPDATED = "firewallrule:sync"
    NETWORK_CONF_UPDATED = "networkconf:sync"
    PING_TEST_UPDATE = "ping-test:update"
    PORT_FORWARD_ADDED = "portforward:add"
    PORT_FORWARD_UPDATED = "portforward:sync"
    PORT_FORWARD_DELETED = "portforward:delete"
    SCHEDULE_TASK_ADDED = "scheduletask:add"
    SESSION_METADATA = "session-metadata:sync"
    SETTING_UPDATED = "setting:sync"
    SETUP_SYNC = "setup:sync"
    SPEED_TEST_UPDATE = "speed-test:update"
    UNIFI_DEVICE = "unifi-device:sync"
    UNIFI_DEVICE_ADD = "unifi-device:add"
    USER_GROUP_UPDATED = "usergroup:sync"
    VPN_CONNECTION_UPDATED = "vpn-connection:sync"
    WLAN_CONF_ADDED = "wlanconf:add"
    WLAN_CONF_UPDATED = "wlanconf:sync"
    WLAN_CONF_DELETED = "wlanconf:delete"

    UNKNOWN = "unknown"

    @classmethod
    def _missing_(cls, value: object) -> "MessageKey":
        """Set default enum member if an unknown value is provided."""
        LOGGER.warning("Unsupported message key %s", value)
        return MessageKey.UNKNOWN


@dataclass
class Meta:
    """Meta description of UniFi websocket data."""

    rc: str
    message: MessageKey
    data: dict[str, Any]

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Self:
        """Create meta instance from dict."""
        return cls(
            rc=data.get("rc", ""),
            message=MessageKey(data.get("message", "")),
            data=data,
        )


@dataclass
class Message:
    """Websocket package representation."""

    meta: Meta
    data: dict[str, Any]

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Self:
        """Create data container instance from dict."""
        meta = Meta.from_dict(data["meta"])
        if meta.message == MessageKey.UNKNOWN:
            LOGGER.warning("Unsupported message %s", data)
        return cls(
            meta=meta,
            data=data["data"],
        )



"""Test API handlers."""

from unittest.mock import Mock

import pytest

from aiounifi.interfaces.api_handlers import APIHandler, ItemEvent


@pytest.mark.parametrize(
    "event_filter",
    [
        None,
        {ItemEvent.ADDED, ItemEvent.CHANGED, ItemEvent.DELETED},
    ],
)
async def test_api_handler_subscriptions(event_filter):
    """Test process and remove item."""
    handler = APIHandler(Mock())
    handler.obj_id_key = "key"
    handler.item_cls = Mock()

    unsub = handler.subscribe(mock_subscribe_cb := Mock(), event_filter)

    handler.process_item({})
    mock_subscribe_cb.assert_not_called()

    handler.process_item({"key": "1"})
    mock_subscribe_cb.assert_called_with(ItemEvent.ADDED, "1")

    handler.process_item({"key": "1"})
    mock_subscribe_cb.assert_called_with(ItemEvent.CHANGED, "1")

    handler.remove_item({"key": "1"})
    mock_subscribe_cb.assert_called_with(ItemEvent.DELETED, "1")

    handler.remove_item({"key": "2"})

    # Process raw

    handler.process_raw([{"key": "2"}])
    mock_subscribe_cb.assert_called_with(ItemEvent.ADDED, "2")

    handler.process_raw([{"key": "2"}])
    mock_subscribe_cb.assert_called_with(ItemEvent.CHANGED, "2")

    handler.remove_item({"key": "2"})
    mock_subscribe_cb.assert_called_with(ItemEvent.DELETED, "2")

    unsub()

    unsub()  # Empty list of object ID

    handler._subscribers.clear()

    unsub()  # Object ID does not exist in subscribers


async def test_api_handler_subscriptions_event_filter_added():
    """Test process and remove item."""
    handler = APIHandler(Mock())
    handler.obj_id_key = "key"
    handler.item_cls = Mock()

    unsub = handler.subscribe(mock_subscribe_cb := Mock(), ItemEvent.ADDED)

    handler.process_item({})
    mock_subscribe_cb.assert_not_called()

    handler.process_item({"key": "1"})
    mock_subscribe_cb.assert_called_with(ItemEvent.ADDED, "1")

    handler.process_item({"key": "1"})
    assert mock_subscribe_cb.call_count == 1

    handler.remove_item({"key": "1"})
    assert mock_subscribe_cb.call_count == 1

    handler.remove_item({"key": "2"})

    # Process raw

    handler.process_raw([{"key": "2"}])
    mock_subscribe_cb.assert_called_with(ItemEvent.ADDED, "2")

    handler.process_raw([{"key": "2"}])
    assert mock_subscribe_cb.call_count == 2

    handler.remove_item({"key": "2"})
    assert mock_subscribe_cb.call_count == 2

    unsub()


async def test_api_handler_subscriptions_id_filter():
    """Test process and remove item."""
    handler = APIHandler(Mock())
    handler.obj_id_key = "key"
    handler.item_cls = Mock()

    unsub = handler.subscribe(mock_subscribe_cb := Mock(), id_filter="1")

    handler.process_item({})
    mock_subscribe_cb.assert_not_called()

    handler.process_item({"key": "1"})
    mock_subscribe_cb.assert_called_with(ItemEvent.ADDED, "1")

    handler.process_item({"key": "1"})
    mock_subscribe_cb.assert_called_with(ItemEvent.CHANGED, "1")

    handler.remove_item({"key": "1"})
    mock_subscribe_cb.assert_called_with(ItemEvent.DELETED, "1")

    handler.remove_item({"key": "2"})

    # Process raw

    handler.process_raw([{"key": "2"}])
    assert mock_subscribe_cb.call_count == 3

    handler.process_raw([{"key": "2"}])
    assert mock_subscribe_cb.call_count == 3

    handler.remove_item({"key": "2"})
    assert mock_subscribe_cb.call_count == 3

    unsub()



"""DPI Restrictions as part of a UniFi network."""

from dataclasses import dataclass
from typing import NotRequired, Self, TypedDict

from .api import ApiItem, ApiRequest


@dataclass
class DpiRestrictionGroupListRequest(ApiRequest):
    """Request object for DPI restriction group list."""

    @classmethod
    def create(cls) -> Self:
        """Create DPI restriction group list request."""
        return cls(method="get", path="/rest/dpigroup")


class TypedDPIRestrictionGroup(TypedDict):
    """DPI restriction group type definition."""

    _id: str
    attr_no_delete: NotRequired[bool]
    attr_hidden_id: str
    name: str
    site_id: str
    dpiapp_ids: NotRequired[list[str]]


class DPIRestrictionGroup(ApiItem):
    """Represents a DPI Group configuration."""

    raw: TypedDPIRestrictionGroup

    @property
    def id(self) -> str:
        """DPI group ID."""
        return self.raw["_id"]

    @property
    def attr_no_delete(self) -> bool | None:
        """Can be deleted."""
        return self.raw.get("attr_no_delete")

    @property
    def attr_hidden_id(self) -> str:
        """Attr hidden ID."""
        return self.raw.get("attr_hidden_id", "")

    @property
    def name(self) -> str:
        """DPI group name."""
        return self.raw["name"]

    @property
    def site_id(self) -> str:
        """Site ID."""
        return self.raw["site_id"]

    @property
    def dpiapp_ids(self) -> list[str] | None:
        """DPI app IDs belonging to group."""
        return self.raw.get("dpiapp_ids")



"""Traffic routes as part of a UniFi network."""

from copy import deepcopy

from ..models.api import TypedApiResponse
from ..models.traffic_route import (
    TrafficRoute,
    TrafficRouteListRequest,
    TrafficRouteSaveRequest,
)
from .api_handlers import APIHandler


class TrafficRoutes(APIHandler[TrafficRoute]):
    """Represents TrafficRoutes configurations."""

    obj_id_key = "_id"
    item_cls = TrafficRoute
    api_request = TrafficRouteListRequest.create()

    async def enable(self, traffic_route: TrafficRoute) -> TypedApiResponse:
        """Enable traffic route defined in controller."""
        return await self.save(traffic_route, state=True)

    async def disable(self, traffic_route: TrafficRoute) -> TypedApiResponse:
        """Disable traffic route defined in controller."""
        return await self.save(traffic_route, state=False)

    async def save(
        self, traffic_route: TrafficRoute, state: bool | None = None
    ) -> TypedApiResponse:
        """Set traffic route - defined in controller - to the desired state."""
        traffic_route_dict = deepcopy(traffic_route.raw)
        traffic_route_response = await self.controller.request(
            TrafficRouteSaveRequest.create(traffic_route_dict, enable=state)
        )
        self.process_raw(traffic_route_response.get("data", []))
        return traffic_route_response



"""Test DPI configuration API.

pytest --cov-report term-missing --cov=aiounifi.dpi tests/test_dpi.py
"""
from collections.abc import Callable
from typing import Any
from unittest.mock import Mock

from aioresponses import aioresponses
import pytest

from aiounifi.controller import Controller
from aiounifi.models.dpi_restriction_app import DPIRestrictionApp
from aiounifi.models.dpi_restriction_group import DPIRestrictionGroup

from .fixtures import DPI_APPS, DPI_GROUPS


@pytest.mark.parametrize("dpi_app_payload", [DPI_APPS])
async def test_dpi_apps(
    mock_aioresponse: aioresponses,
    unifi_controller: Controller,
    _mock_endpoints: None,
    unifi_called_with: Callable[[str, str, dict[str, Any]], bool],
) -> None:
    """Test that dpi_apps can create an app."""
    dpi_apps = unifi_controller.dpi_apps
    await dpi_apps.update()
    assert len(dpi_apps.values()) == 1

    app: DPIRestrictionApp = dpi_apps["5f976f62e3c58f018ec7e17d"]
    assert app.id == "5f976f62e3c58f018ec7e17d"
    assert app.apps == []
    assert app.blocked
    assert app.cats == ["4"]
    assert app.enabled
    assert app.log
    assert app.site_id == "5ba29dd4e3c58f026e9d7c38"

    mock_aioresponse.put(
        "https://host:8443/api/s/default/rest/dpiapp/5f976f62e3c58f018ec7e17d",
        payload={},
        repeat=True,
    )
    await dpi_apps.enable("5f976f62e3c58f018ec7e17d")
    assert unifi_called_with(
        "put",
        "/api/s/default/rest/dpiapp/5f976f62e3c58f018ec7e17d",
        json={"enabled": True},
    )

    await dpi_apps.disable("5f976f62e3c58f018ec7e17d")
    assert unifi_called_with(
        "put",
        "/api/s/default/rest/dpiapp/5f976f62e3c58f018ec7e17d",
        json={"enabled": False},
    )


@pytest.mark.parametrize("dpi_group_payload", [DPI_GROUPS])
async def test_dpi_groups(unifi_controller: Controller, _mock_endpoints: None) -> None:
    """Test that dpi_groups can create a group."""
    dpi_groups = unifi_controller.dpi_groups
    await dpi_groups.update()
    assert len(dpi_groups.values()) == 2

    group: DPIRestrictionGroup = dpi_groups["5f976f4ae3c58f018ec7dff6"]
    assert group.id == "5f976f4ae3c58f018ec7dff6"
    assert not group.attr_no_delete
    assert group.attr_hidden_id == ""
    assert group.name == "No Media"
    assert group.site_id == "5ba29dd4e3c58f026e9d7c38"
    assert group.dpiapp_ids == ["5f976f62e3c58f018ec7e17d"]


async def test_dpi_apps_websocket(
    unifi_controller: Controller, _new_ws_data_fn: Callable[[dict[str, Any]], None]
) -> None:
    """Test controller managing devices."""
    unifi_controller.dpi_apps.subscribe(mock_app_callback := Mock())

    # Add DPI app from websocket
    _new_ws_data_fn(
        {
            "meta": {"rc": "ok", "message": "dpiapp:add"},
            "data": [
                {
                    "apps": [524292],
                    "blocked": False,
                    "cats": [],
                    "enabled": False,
                    "log": False,
                    "site_id": "5f3edd27ba4cc806a19f2d9c",
                    "_id": "61783e89c1773a18c0c61f00",
                }
            ],
        }
    )
    assert len(unifi_controller.dpi_apps.values()) == 1
    assert "61783e89c1773a18c0c61f00" in unifi_controller.dpi_apps

    mock_app_callback.assert_called()
    mock_app_callback.reset_mock()

    # DPI group is enabled with app from websocket
    _new_ws_data_fn(
        {
            "meta": {"rc": "ok", "message": "dpiapp:sync"},
            "data": [
                {
                    "_id": "61783e89c1773a18c0c61f00",
                    "apps": [524292],
                    "blocked": False,
                    "cats": [],
                    "enabled": True,
                    "log": False,
                    "site_id": "5f3edd27ba4cc806a19f2d9c",
                }
            ],
        }
    )
    dpi_app = unifi_controller.dpi_apps["61783e89c1773a18c0c61f00"]
    assert dpi_app.enabled

    mock_app_callback.assert_called()
    mock_app_callback.reset_mock()

    # Signal removal of app from apps
    _new_ws_data_fn(
        {
            "meta": {"rc": "ok", "message": "dpiapp:delete"},
            "data": [
                {
                    "_id": "61783e89c1773a18c0c61f00",
                    "apps": [524292],
                    "blocked": False,
                    "cats": [],
                    "enabled": True,
                    "log": False,
                    "site_id": "5f3edd27ba4cc806a19f2d9c",
                }
            ],
        }
    )
    assert len(unifi_controller.dpi_apps.values()) == 0
    assert "61783e89c1773a18c0c61f00" not in unifi_controller.dpi_apps
    mock_app_callback.assert_called()


async def test_dpi_group_websocket(
    unifi_controller: Controller, _new_ws_data_fn: Callable[[dict[str, Any]], None]
) -> None:
    """Test controller managing devices."""
    unifi_controller.dpi_groups.subscribe(mock_group_callback := Mock())

    # Add DPI group from websocket
    _new_ws_data_fn(
        {
            "meta": {"rc": "ok", "message": "dpigroup:add"},
            "data": [
                {
                    "name": "dpi group",
                    "site_id": "5f3edd27ba4cc806a19f2d9c",
                    "_id": "61783dbdc1773a18c0c61ef6",
                }
            ],
        }
    )
    assert len(unifi_controller.dpi_groups.values()) == 1
    assert "61783dbdc1773a18c0c61ef6" in unifi_controller.dpi_groups

    mock_group_callback.assert_called()
    mock_group_callback.reset_mock()

    # Update DPI group with app from websocket
    _new_ws_data_fn(
        {
            "meta": {"rc": "ok", "message": "dpigroup:sync"},
            "data": [
                {
                    "_id": "61783dbdc1773a18c0c61ef6",
                    "name": "dpi group",
                    "site_id": "5f3edd27ba4cc806a19f2d9c",
                    "dpiapp_ids": ["61783e89c1773a18c0c61f00"],
                }
            ],
        }
    )
    dpi_group = unifi_controller.dpi_groups["61783dbdc1773a18c0c61ef6"]
    assert "61783e89c1773a18c0c61f00" in dpi_group.dpiapp_ids

    mock_group_callback.assert_called()
    mock_group_callback.reset_mock()

    # Signal for group to remove app
    _new_ws_data_fn(
        {
            "meta": {"rc": "ok", "message": "dpigroup:sync"},
            "data": [
                {
                    "_id": "61783dbdc1773a18c0c61ef6",
                    "name": "dpi group",
                    "site_id": "5f3edd27ba4cc806a19f2d9c",
                    "dpiapp_ids": [],
                }
            ],
        }
    )
    mock_group_callback.assert_called()
    mock_group_callback.reset_mock()

    # Remove group from UniFI controller group from websocket
    _new_ws_data_fn(
        {
            "meta": {"rc": "ok", "message": "dpigroup:delete"},
            "data": [
                {
                    "_id": "61783dbdc1773a18c0c61ef6",
                    "name": "dpi group",
                    "site_id": "5f3edd27ba4cc806a19f2d9c",
                    "dpiapp_ids": [],
                }
            ],
        }
    )
    mock_group_callback.assert_called()
    assert len(unifi_controller.dpi_groups.values()) == 0
    assert "61783dbdc1773a18c0c61ef6" not in unifi_controller.dpi_groups



"""Test client configuration API.

pytest --cov-report term-missing --cov=aiounifi.clients tests/test_clients.py
"""
from collections.abc import Callable
from typing import Any
from unittest.mock import Mock

from aioresponses import aioresponses
import pytest

from aiounifi.controller import Controller
from aiounifi.models.message import MessageKey

from .fixtures import MESSAGE_WIRELESS_CLIENT_REMOVED, WIRED_CLIENT, WIRELESS_CLIENT

test_data = [
    (
        [{"mac": "0"}],
        {
            "access_point_mac": "",
            "association_time": 0,
            "blocked": False,
            "device_name": "",
            "essid": "",
            "firmware_version": "",
            "first_seen": 0,
            "fixed_ip": "",
            "hostname": "",
            "idle_time": 0,
            "ip": "",
            "is_guest": False,
            "is_wired": False,
            "last_seen": 0,
            "last_seen_by_access_point": 0,
            "last_seen_by_gateway": 0,
            "last_seen_by_switch": 0,
            "latest_association_time": 0,
            "mac": "0",
            "name": "",
            "oui": "",
            "powersave_enabled": None,
            "site_id": "",
            "switch_depth": None,
            "switch_mac": "",
            "switch_port": None,
            "rx_bytes": 0,
            "rx_bytes_r": 0,
            "tx_bytes": 0,
            "tx_bytes_r": 0,
            "uptime": 0,
            "uptime_by_access_point": 0,
            "uptime_by_gateway": 0,
            "uptime_by_switch": 0,
            "wired_rate_mbps": 0,
            "wired_rx_bytes": 0,
            "wired_rx_bytes_r": 0,
            "wired_tx_bytes": 0,
            "wired_tx_bytes_r": 0,
        },
    ),
    (
        [WIRELESS_CLIENT],
        {
            "access_point_mac": "80:2a:a8:00:01:02",
            "association_time": 1587753456,
            "blocked": False,
            "device_name": "Discovery device name",
            "essid": "SSID",
            "firmware_version": "",
            "first_seen": 1513271497,
            "fixed_ip": "192.168.0.1",
            "hostname": "client",
            "idle_time": 1,
            "ip": "192.168.0.1",
            "is_guest": False,
            "is_wired": False,
            "last_seen": 1587765360,
            "last_seen_by_access_point": 1587765360,
            "last_seen_by_gateway": 1587765372,
            "last_seen_by_switch": 1587763868,
            "latest_association_time": 1587765354,
            "mac": WIRELESS_CLIENT["mac"],
            "name": "Client 1",
            "oui": "Apple",
            "powersave_enabled": False,
            "site_id": "5a32aa4ee4b0412345678910",
            "switch_depth": -1,
            "switch_mac": "fc:ec:da:11:22:33",
            "switch_port": 1,
            "rx_bytes": 12867114,
            "rx_bytes_r": 326,
            "tx_bytes": 52852089,
            "tx_bytes_r": 483,
            "uptime": 11904,
            "uptime_by_access_point": 11904,
            "uptime_by_gateway": 18,
            "uptime_by_switch": 318,
            "wired_rate_mbps": 0,
            "wired_rx_bytes": 0,
            "wired_rx_bytes_r": 0,
            "wired_tx_bytes": 0,
            "wired_tx_bytes_r": 0,
        },
    ),
    (
        [WIRED_CLIENT],
        {
            "access_point_mac": "",
            "association_time": 1642487042,
            "blocked": False,
            "device_name": "Network Camera",
            "essid": "",
            "firmware_version": "",
            "first_seen": 1598100273,
            "fixed_ip": "",
            "hostname": "camera-acccde123456",
            "idle_time": 0,
            "ip": "192.168.0.2",
            "is_guest": False,
            "is_wired": True,
            "last_seen": 1642574376,
            "last_seen_by_access_point": 0,
            "last_seen_by_gateway": 1642574376,
            "last_seen_by_switch": 1642574376,
            "latest_association_time": 1642487042,
            "mac": WIRED_CLIENT["mac"],
            "name": "",
            "oui": "Manu",
            "powersave_enabled": None,
            "site_id": "5a32aa4ee4b0412345678910",
            "switch_depth": 0,
            "switch_mac": "fc:ec:da:11:22:33",
            "switch_port": 3,
            "rx_bytes": 0,
            "rx_bytes_r": 0,
            "tx_bytes": 0,
            "tx_bytes_r": 0,
            "uptime": 87334,
            "uptime_by_access_point": 0,
            "uptime_by_gateway": 1527337,
            "uptime_by_switch": 1527337,
            "wired_rate_mbps": 1000,
            "wired_rx_bytes": 0,
            "wired_rx_bytes_r": 0,
            "wired_tx_bytes": 0,
            "wired_tx_bytes_r": 0,
        },
    ),
]


@pytest.mark.parametrize(("client_payload", "reference_data"), test_data)
async def test_clients(
    unifi_controller: Controller, _mock_endpoints: None, reference_data: dict[str, Any]
) -> None:
    """Test clients class."""
    clients = unifi_controller.clients
    await clients.update()
    assert len(clients.items()) == 1

    client = next(iter(clients.values()))
    for key, value in reference_data.items():
        assert getattr(client, key) == value


@pytest.mark.parametrize(
    ("method", "mac", "command"),
    [
        ["block", "0", {"mac": "0", "cmd": "block-sta"}],
        ["unblock", "0", {"mac": "0", "cmd": "unblock-sta"}],
        ["reconnect", "0", {"mac": "0", "cmd": "kick-sta"}],
        ["remove_clients", ["0"], {"macs": ["0"], "cmd": "forget-sta"}],
    ],
)
async def test_client_commands(
    mock_aioresponse: aioresponses,
    unifi_controller: Controller,
    unifi_called_with: Callable[[str, str, dict[str, Any]], bool],
    method: str,
    mac: str | list[str],
    command: dict[str, str | list[str]],
) -> None:
    """Test client commands."""
    mock_aioresponse.post("https://host:8443/api/s/default/cmd/stamgr", payload={})
    class_command = getattr(unifi_controller.clients, method)
    await class_command(mac)
    assert unifi_called_with("post", "/api/s/default/cmd/stamgr", json=command)


async def test_client_websocket(
    unifi_controller: Controller, _new_ws_data_fn: Callable[[dict[str, Any]], None]
) -> None:
    """Test controller managing clients."""
    unsub = unifi_controller.clients.subscribe(mock_callback := Mock())
    assert len(unifi_controller.clients._subscribers["*"]) == 1
    assert mock_callback.call_count == 0

    # Add client from websocket
    _new_ws_data_fn(
        {
            "meta": {"message": MessageKey.CLIENT.value},
            "data": [WIRELESS_CLIENT],
        }
    )
    assert len(unifi_controller.clients.items()) == 1
    assert mock_callback.call_count == 1

    # Remove callback
    unsub()
    assert len(unifi_controller.clients._subscribers["*"]) == 0


@pytest.mark.parametrize("client_payload", [[WIRELESS_CLIENT]])
async def test_message_client_removed(
    unifi_controller: Controller,
    _mock_endpoints: None,
    _new_ws_data_fn: Callable[[dict[str, Any]], None],
) -> None:
    """Test controller communicating client has been removed."""
    await unifi_controller.initialize()
    assert len(unifi_controller.clients.items()) == 1

    _new_ws_data_fn(MESSAGE_WIRELESS_CLIENT_REMOVED)
    assert len(unifi_controller.clients.items()) == 0



"""UniFi system information model."""

from dataclasses import dataclass
from typing import TypedDict

from .api import ApiItem, ApiRequest


class TypedSystemInfo(TypedDict):
    """System information type definition."""

    anonymous_controller_id: str
    autobackup: bool
    build: str
    console_display_version: str
    data_retention_days: int
    data_retention_time_in_hours_for_5minutes_scale: int
    data_retention_time_in_hours_for_daily_scale: int
    data_retention_time_in_hours_for_hourly_scale: int
    data_retention_time_in_hours_for_monthly_scale: int
    data_retention_time_in_hours_for_others: int
    debug_device: str
    debug_mgmt: str
    debug_sdn: str
    debug_setting_preference: str
    debug_system: str
    default_site_device_auth_password_alert: bool
    facebook_wifi_registered: bool
    has_webrtc_support: bool
    hostname: str
    https_port: int
    image_maps_use_google_engine: bool
    inform_port: int
    ip_addrs: list[str]
    is_cloud_console: bool
    live_chat: str
    name: str
    override_inform_host: bool
    portal_http_port: int
    previous_version: str
    radius_disconnect_running: bool
    sso_app_id: str
    sso_app_sec: str
    store_enabled: str
    timezone: str
    ubnt_device_type: str
    udm_version: str
    unifi_go_enabled: bool
    unsupported_device_count: int
    unsupported_device_list: list[str]
    update_available: bool
    update_downloaded: bool
    uptime: int
    version: str


@dataclass
class SystemInformationRequest(ApiRequest):
    """Request object for system information."""

    @classmethod
    def create(cls) -> "SystemInformationRequest":
        """Create system information request."""
        return cls(method="get", path="/stat/sysinfo")


class SystemInformation(ApiItem):
    """Represents a client network device."""

    raw: TypedSystemInfo

    @property
    def anonymous_controller_id(self) -> str:
        """Anonymous controller ID."""
        return self.raw["anonymous_controller_id"]

    @property
    def device_type(self) -> str:
        """Network host device type."""
        return self.raw["ubnt_device_type"]

    @property
    def hostname(self) -> str:
        """Host name."""
        return self.raw["hostname"]

    @property
    def ip_address(self) -> list[str]:
        """External IP address."""
        return self.raw["ip_addrs"]

    @property
    def is_cloud_console(self) -> bool:
        """Cloud hosted console."""
        return self.raw["is_cloud_console"]

    @property
    def name(self) -> str:
        """Name."""
        return self.raw["name"]

    @property
    def previous_version(self) -> str:
        """Previous application version."""
        return self.raw["previous_version"]

    @property
    def update_available(self) -> bool:
        """Application update available."""
        return self.raw["update_available"]

    @property
    def uptime(self) -> int:
        """Application uptime."""
        return self.raw["uptime"]

    @property
    def version(self) -> str:
        """Current application version."""
        return self.raw["version"]



"""WLANs as part of a UniFi network."""

from dataclasses import dataclass
import io
from typing import NotRequired, Self, TypedDict

import segno.helpers

from .api import ApiItem, ApiRequest


class TypedWlan(TypedDict):
    """Wlan type definition."""

    _id: str
    bc_filter_enabled: bool
    bc_filter_list: list[str]
    dtim_mode: str
    dtim_na: int
    dtim_ng: int
    enabled: bool
    group_rekey: int
    is_guest: NotRequired[bool]
    mac_filter_enabled: NotRequired[bool]
    mac_filter_list: list[str]
    mac_filter_policy: str
    minrate_na_advertising_rates: bool
    minrate_na_beacon_rate_kbps: int
    minrate_na_data_rate_kbps: int
    minrate_na_enabled: NotRequired[bool]
    minrate_na_mgmt_rate_kbps: int
    minrate_ng_advertising_rates: bool
    minrate_ng_beacon_rate_kbps: int
    minrate_ng_cck_rates_enabled: NotRequired[bool]
    minrate_ng_data_rate_kbps: int
    minrate_ng_enabled: NotRequired[bool]
    minrate_ng_mgmt_rate_kbps: int
    name: str
    name_combine_enabled: NotRequired[bool]
    name_combine_suffix: NotRequired[str]
    no2ghz_oui: bool
    schedule: list[str]
    security: str
    site_id: str
    usergroup_id: str
    wep_idx: int
    wlangroup_id: str
    wpa_enc: str
    wpa_mode: str
    x_iapp_key: str
    x_passphrase: NotRequired[str]


@dataclass
class WlanListRequest(ApiRequest):
    """Request object for wlan list."""

    @classmethod
    def create(cls) -> Self:
        """Create wlan list request."""
        return cls(method="get", path="/rest/wlanconf")


@dataclass
class WlanChangePasswordRequest(ApiRequest):
    """Request object for wlan password change."""

    @classmethod
    def create(cls, wlan_id: str, password: str) -> Self:
        """Create wlan password change request."""
        return cls(
            method="put",
            path=f"/rest/wlanconf/{wlan_id}",
            data={"x_passphrase": password},
        )


@dataclass
class WlanEnableRequest(ApiRequest):
    """Request object for wlan enable."""

    @classmethod
    def create(cls, wlan_id: str, enable: bool) -> Self:
        """Create wlan enable request."""
        return cls(
            method="put",
            path=f"/rest/wlanconf/{wlan_id}",
            data={"enabled": enable},
        )


def wlan_qr_code(
    name: str, password: str | None, kind: str = "png", scale: int = 4
) -> bytes:
    """Generate WLAN QR code."""
    buffer = io.BytesIO()
    qr_code = segno.helpers.make_wifi(ssid=name, password=password, security="WPA")
    qr_code.save(out=buffer, kind=kind, scale=scale)
    return buffer.getvalue()


class Wlan(ApiItem):
    """Represent a WLAN configuration."""

    raw: TypedWlan

    @property
    def id(self) -> str:
        """ID of WLAN."""
        return self.raw["_id"]

    @property
    def bc_filter_enabled(self) -> bool:
        """Is BC filter enabled."""
        return self.raw.get("bc_filter_enabled", False)

    @property
    def bc_filter_list(self) -> list[str]:
        """List of BC filters."""
        return self.raw["bc_filter_list"]

    @property
    def dtim_mode(self) -> str:
        """DTIM mode."""
        return self.raw["dtim_mode"]

    @property
    def dtim_na(self) -> int:
        """DTIM NA."""
        return self.raw["dtim_na"]

    @property
    def dtim_ng(self) -> int:
        """DTIM NG."""
        return self.raw["dtim_ng"]

    @property
    def enabled(self) -> bool:
        """Is WLAN enabled."""
        return self.raw["enabled"]

    @property
    def group_rekey(self) -> int:
        """Group rekey."""
        return self.raw["group_rekey"]

    @property
    def is_guest(self) -> bool | None:
        """Is WLAN a guest network."""
        return self.raw.get("is_guest")

    @property
    def mac_filter_enabled(self) -> bool | None:
        """Is MAC filtering enabled."""
        return self.raw.get("mac_filter_enabled")

    @property
    def mac_filter_list(self) -> list[str]:
        """List of MAC filters."""
        return self.raw["mac_filter_list"]

    @property
    def mac_filter_policy(self) -> str:
        """MAC filter policy."""
        return self.raw["mac_filter_policy"]

    @property
    def minrate_na_advertising_rates(self) -> bool:
        """Minrate NA advertising rates."""
        return self.raw["minrate_na_advertising_rates"]

    @property
    def minrate_na_beacon_rate_kbps(self) -> int:
        """Minrate NA beacon rate kbps."""
        return self.raw["minrate_na_beacon_rate_kbps"]

    @property
    def minrate_na_data_rate_kbps(self) -> int:
        """Minrate NA data rate kbps."""
        return self.raw["minrate_na_data_rate_kbps"]

    @property
    def minrate_na_enabled(self) -> bool | None:
        """Is minrate NA enabled."""
        return self.raw.get("minrate_na_enabled")

    @property
    def minrate_na_mgmt_rate_kbps(self) -> int:
        """Minrate NA management rate kbps."""
        return self.raw["minrate_na_mgmt_rate_kbps"]

    @property
    def minrate_ng_advertising_rates(self) -> bool:
        """Minrate NG advertising rates."""
        return self.raw["minrate_ng_advertising_rates"]

    @property
    def minrate_ng_beacon_rate_kbps(self) -> int:
        """Minrate NG beacon rate kbps."""
        return self.raw["minrate_ng_beacon_rate_kbps"]

    @property
    def minrate_ng_cck_rates_enabled(self) -> bool | None:
        """Is minrate NG CCK rates enabled."""
        return self.raw.get("minrate_ng_cck_rates_enabled")

    @property
    def minrate_ng_data_rate_kbps(self) -> int:
        """Minrate NG data rate kbps."""
        return self.raw["minrate_ng_data_rate_kbps"]

    @property
    def minrate_ng_enabled(self) -> bool | None:
        """Is minrate NG enabled."""
        return self.raw.get("minrate_ng_enabled")

    @property
    def minrate_ng_mgmt_rate_kbps(self) -> int:
        """Minrate NG management rate kbps."""
        return self.raw["minrate_ng_mgmt_rate_kbps"]

    @property
    def name(self) -> str:
        """WLAN name."""
        return self.raw["name"]

    @property
    def name_combine_enabled(self) -> bool | None:
        """If 2.5 and 5 GHz SSIDs should be combined."""
        return self.raw.get("name_combine_enabled")

    @property
    def name_combine_suffix(self) -> str | None:
        """Suffix for 2.4GHz SSID if name is not combined."""
        return self.raw.get("name_combine_suffix")

    @property
    def no2ghz_oui(self) -> bool:
        """No 2GHz OUI."""
        return self.raw["no2ghz_oui"]

    @property
    def schedule(self) -> list[str]:
        """Schedule list."""
        return self.raw["schedule"]

    @property
    def security(self) -> str:
        """Security."""
        return self.raw["security"]

    @property
    def site_id(self) -> str:
        """WLAN site ID."""
        return self.raw["site_id"]

    @property
    def usergroup_id(self) -> str:
        """WLAN user group ID."""
        return self.raw["usergroup_id"]

    @property
    def wep_idx(self) -> int:
        """WEP index."""
        return self.raw["wep_idx"]

    @property
    def wlangroup_id(self) -> str:
        """WLAN group ID."""
        return self.raw["wlangroup_id"]

    @property
    def wpa_enc(self) -> str:
        """WPA encryption."""
        return self.raw["wpa_enc"]

    @property
    def wpa_mode(self) -> str:
        """WPA mode."""
        return self.raw["wpa_mode"]

    @property
    def x_iapp_key(self) -> str:
        """X iapp key."""
        return self.raw["x_iapp_key"]

    @property
    def x_passphrase(self) -> str | None:
        """Passphrase."""
        return self.raw.get("x_passphrase")



"""Python library to enable integration between Home Assistant and UniFi."""

from collections.abc import Callable, Mapping
from http import HTTPStatus
import logging
from typing import TYPE_CHECKING, Any

import aiohttp
from aiohttp import client_exceptions
import orjson

from ..errors import (
    AiounifiException,
    BadGateway,
    Forbidden,
    LoginRequired,
    RequestError,
    ResponseError,
    ServiceUnavailable,
    WebsocketError,
)
from ..models.api import ERRORS
from ..models.configuration import Configuration

if TYPE_CHECKING:
    from ..models.api import ApiRequest, TypedApiResponse

LOGGER = logging.getLogger(__name__)


class Connectivity:
    """UniFi Network Application connectivity."""

    def __init__(self, config: Configuration) -> None:
        """Session setup."""
        self.config = config

        self.is_unifi_os = False
        self.headers: dict[str, str] = {}
        self.can_retry_login = False

    async def check_unifi_os(self) -> None:
        """Check if controller is running UniFi OS."""
        response, _ = await self._request("get", self.config.url, allow_redirects=False)
        if response.status == HTTPStatus.OK:
            self.is_unifi_os = True
            self.config.session.cookie_jar.clear_domain(self.config.host)
        LOGGER.debug("Talking to UniFi OS device: %s", self.is_unifi_os)

    async def login(self) -> None:
        """Log in to controller."""
        self.headers.clear()
        url = f"{self.config.url}/api{'/auth/login' if self.is_unifi_os else '/login'}"

        auth = {
            "username": self.config.username,
            "password": self.config.password,
            "remember": True,
        }

        response, bytes_data = await self._request("post", url, json=auth)

        if response.content_type == "application/json":
            data: "TypedApiResponse" = orjson.loads(bytes_data)
            if "meta" in data and data["meta"]["rc"] == "error":
                LOGGER.error("Login failed '%s'", data)
                raise ERRORS.get(data["meta"]["msg"], AiounifiException)

        if (
            response.status == HTTPStatus.OK
            and (csrf_token := response.headers.get("x-csrf-token")) is not None
        ):
            self.headers["x-csrf-token"] = csrf_token

        self.can_retry_login = True
        LOGGER.debug("Logged in to UniFi %s", url)

    async def request(self, api_request: "ApiRequest") -> "TypedApiResponse":
        """Make a request to the API, retry login on failure."""
        url = self.config.url + api_request.full_path(
            self.config.site, self.is_unifi_os
        )
        data: TypedApiResponse = {}

        try:
            response, bytes_data = await self._request(
                api_request.method, url, api_request.data
            )

            if response.content_type == "application/json":
                data = api_request.decode(bytes_data)

        except LoginRequired:
            if not self.can_retry_login:
                raise
            # Session likely expired, try again
            self.can_retry_login = False
            await self.login()
            return await self.request(api_request)

        return data

    async def _request(
        self,
        method: str,
        url: str,
        json: Mapping[str, Any] | None = None,
        **kwargs: bool,
    ) -> tuple[aiohttp.ClientResponse, bytes]:
        """Make a request to the API."""
        LOGGER.debug("sending (to %s) %s, %s, %s", url, method, json, kwargs)
        bytes_data = b""

        try:
            async with self.config.session.request(
                method,
                url,
                json=json,
                ssl=self.config.ssl_context,
                headers=self.headers,
                **kwargs,
            ) as res:
                LOGGER.debug(
                    "received (from %s) %s %s %s",
                    url,
                    res.status,
                    res.content_type,
                    res,
                )

                if res.status == HTTPStatus.UNAUTHORIZED:
                    raise LoginRequired(f"Call {url} received 401 Unauthorized")

                if res.status == HTTPStatus.FORBIDDEN:
                    raise Forbidden(f"Call {url} received 403 Forbidden")

                if res.status == HTTPStatus.NOT_FOUND:
                    raise ResponseError(f"Call {url} received 404 Not Found")

                if res.status == HTTPStatus.BAD_GATEWAY:
                    raise BadGateway(f"Call {url} received 502 bad gateway")

                if res.status == HTTPStatus.SERVICE_UNAVAILABLE:
                    raise ServiceUnavailable(
                        f"Call {url} received 503 service unavailable"
                    )

                bytes_data = await res.read()

        except client_exceptions.ClientError as err:
            raise RequestError(f"Error requesting data from {url}: {err}") from None

        LOGGER.debug("data (from %s) %s", url, bytes_data)
        return res, bytes_data

    async def websocket(self, callback: Callable[[bytes], None]) -> None:
        """Run websocket."""
        url = f"wss://{self.config.host}:{self.config.port}"
        url += "/proxy/network" if self.is_unifi_os else ""
        url += f"/wss/s/{self.config.site}/events"

        try:
            async with self.config.session.ws_connect(
                url, ssl=self.config.ssl_context, heartbeat=15
            ) as websocket_connection:
                LOGGER.debug("Connected to UniFi websocket %s", url)

                async for message in websocket_connection:
                    if message.type == aiohttp.WSMsgType.TEXT:
                        LOGGER.debug("Websocket '%s'", message.data)
                        callback(message.data)

                    elif message.type == aiohttp.WSMsgType.CLOSED:
                        LOGGER.warning(
                            "Connection closed to UniFi websocket '%s'", message.data
                        )
                        break

                    elif message.type == aiohttp.WSMsgType.ERROR:
                        LOGGER.error("UniFi websocket error: '%s'", message.data)
                        raise WebsocketError(message.data)

                    else:
                        LOGGER.warning(
                            "Unexpected websocket message type '%s' with data '%s'",
                            message.type,
                            message.data,
                        )

        except aiohttp.ClientConnectorError as err:
            LOGGER.error("Error connecting to UniFi websocket: '%s'", err)
            err.add_note("Error connecting to UniFi websocket")
            raise

        except WebsocketError:
            raise

        except Exception as err:
            LOGGER.exception(err)
            raise WebsocketError from err



"""Test wlan configuration API.

pytest --cov-report term-missing --cov=aiounifi.wlan tests/test_wlans.py
"""

import pytest

from aiounifi.models.wlan import (
    WlanChangePasswordRequest,
    WlanEnableRequest,
    wlan_qr_code,
)

from .fixtures import WLANS


async def test_wlan_password_change(
    mock_aioresponse, unifi_controller, unifi_called_with
):
    """Test changing Wlan password."""
    mock_aioresponse.put(
        "https://host:8443/api/s/default/rest/wlanconf/123", payload={}
    )

    await unifi_controller.request(WlanChangePasswordRequest.create("123", "456"))

    assert unifi_called_with(
        "put",
        "/api/s/default/rest/wlanconf/123",
        json={"x_passphrase": "456"},
    )


@pytest.mark.parametrize("enable", [True, False])
async def test_wlan_enable(
    mock_aioresponse, unifi_controller, unifi_called_with, enable
):
    """Test that wlan can be enabled and disabled."""
    mock_aioresponse.put(
        "https://host:8443/api/s/default/rest/wlanconf/123", payload={}
    )

    await unifi_controller.request(WlanEnableRequest.create("123", enable))

    assert unifi_called_with(
        "put",
        "/api/s/default/rest/wlanconf/123",
        json={"enabled": enable},
    )


def test_wlan_qr_code():
    """Test that wlan can be enabled and disabled."""
    assert (
        wlan_qr_code("ssid", "passphrase")
        == b"\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR\x00\x00\x00\x84\x00\x00\x00"
        + b"\x84\x01\x00\x00\x00\x00y?\xbe\n\x00\x00\x00\xc4IDATx\xda\xedV\xcb"
        + b"\r\xc3P\x0cB]\xc0\xfbo\xc9\x06\x14\xfcri\x0fUUz\x8c\xf3\xd1\x93\xa5"
        + b" \x8c\xb1\x15\xe8-\x88;\xf3\x9a\x010\x1cir\x10\x1fx\x8fo2\xc6\t\x84"
        + b"\x86\xd7\xa1\xc0\xc1\x0c\x91\xb7\xef\x12\x07\xe1\xf2\x07\x9c\x81\xa6"
        + b"\xc6\xd1\xaac\xa0N\x9f\xb4i\xa3\xec\xd7F\xc8\x1c'\xfc\xce\xe7\xb8H"
        + b"\xe7)p\\LB\xe8t>U\x81\xeb\xa2\x02\x87\x80\xb9\x0c\x99\xbb\xe9;\x07V"
        + b"\xc8 \x8bZ\xe8c\xf7D\x1djJ>\t\xee\xd5\xe0\xf8\xe3\xdd\x1a.L\xd5|\xb9M{"
        + b"\x9e\xda\xcff\xe2\x9a:\x9c%\xc2k\xe8\xcby\xdf\xa5!\xf4\xfb\xd0^\x8eL5"
        + b"\x0e\xd7\xd7\xa3\x9e\xcf\x8eG\xabO\xbcc\x8d\xfa}\x98\xbe\xb3\xdb\xab"
        + b"\xf7_\xc1\x87\xcc\x13u:\xfe\x00\xeaq\xcby\x00\x00\x00\x00IEND\xaeB`\x82"
    )


async def test_no_wlans(unifi_controller, _mock_endpoints, unifi_called_with):
    """Test that no ports also work."""
    wlans = unifi_controller.wlans
    await wlans.update()

    assert unifi_called_with("get", "/api/s/default/rest/wlanconf")
    assert len(wlans.values()) == 0


@pytest.mark.parametrize("wlan_payload", [WLANS])
async def test_wlans(
    mock_aioresponse, unifi_controller, _mock_endpoints, unifi_called_with
):
    """Test that different types of ports work."""
    wlans = unifi_controller.wlans
    await wlans.update()
    assert len(wlans.values()) == 2

    wlan = wlans["012345678910111213141516"]
    assert wlan.id == "012345678910111213141516"
    assert wlan.bc_filter_enabled is False
    assert wlan.bc_filter_list == []
    assert wlan.dtim_mode == "default"
    assert wlan.dtim_na == 1
    assert wlan.dtim_ng == 1
    assert wlan.enabled is True
    assert wlan.group_rekey == 3600
    assert wlan.is_guest is None
    assert wlan.mac_filter_enabled is False
    assert wlan.mac_filter_list == []
    assert wlan.mac_filter_policy == "allow"
    assert wlan.minrate_na_advertising_rates is False
    assert wlan.minrate_na_beacon_rate_kbps == 6000
    assert wlan.minrate_na_data_rate_kbps == 6000
    assert wlan.minrate_na_enabled is False
    assert wlan.minrate_na_mgmt_rate_kbps == 6000
    assert wlan.minrate_ng_advertising_rates is False
    assert wlan.minrate_ng_beacon_rate_kbps == 1000
    assert wlan.minrate_ng_cck_rates_enabled is None
    assert wlan.minrate_ng_data_rate_kbps == 1000
    assert wlan.minrate_ng_enabled is False
    assert wlan.minrate_ng_mgmt_rate_kbps == 1000
    assert wlan.name == "SSID 1"
    assert wlan.name_combine_enabled is None
    assert wlan.name_combine_suffix is None
    assert wlan.no2ghz_oui is False
    assert wlan.schedule == []
    assert wlan.security == "wpapsk"
    assert wlan.site_id == "5a32aa4ee4b0412345678910"
    assert wlan.usergroup_id == "012345678910111213141518"
    assert wlan.wep_idx == 1
    assert wlan.wlangroup_id == "012345678910111213141519"
    assert wlan.wpa_enc == "ccmp"
    assert wlan.wpa_mode == "wpa2"
    assert wlan.x_iapp_key == "01234567891011121314151617181920"
    assert wlan.x_passphrase == "password in clear text"

    mock_aioresponse.put(
        "https://host:8443/api/s/default/rest/wlanconf/012345678910111213141516",
        payload={},
        repeat=True,
    )

    await wlans.enable(wlan)
    assert unifi_called_with(
        "put",
        "/api/s/default/rest/wlanconf/012345678910111213141516",
        json={"enabled": True},
    )

    await wlans.disable(wlan)
    assert unifi_called_with(
        "put",
        "/api/s/default/rest/wlanconf/012345678910111213141516",
        json={"enabled": False},
    )

    assert (
        wlans.generate_wlan_qr_code(wlan)
        == b"\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR\x00\x00\x00\x94\x00\x00\x00"
        + b"\x94\x01\x00\x00\x00\x00]G=y\x00\x00\x00\xf6IDATx\xda\xed\x96A\x0eBA"
        + b"\x08C\x89\x17\xe0\xfe\xb7\xec\r\xb0\x0f\xbf+7\xeat\xf9\tF2\x8b\xa6C"
        + b"\x99\xf2k>Bu\x9f\xfdrVU-\xd7\xad\xa6\x1c=\xea3\xbe=3^;\xcbX.(\xcf\xf1"
        + b"\xaa\xdb<\x0bD\x97\x11\xbc\x92\xda\x19\xc3\x1b2\xc5\xcf\tL\x87\xfa\x87"
        + b"\xa8WD\xf4}\x85\xd9\xbdg\xe8\x98_oj\xac\x8a\xce\xf1\x04\xd2\x12\xe4\xc6"
        + b"\x13\xb8\xaf\x87YL\x0c\x1c\x13\xfc\x9c\xfe\r\x03\xa8\x80\x1e\xed\x16"
        + b"\xc2\xcf#my\xcf\xf9\xb9i\x0c\x1f\x0c\x13\xfd\x83\x95\x83\xfei\x02\xfc"
        + b"\x06\x98\xebq$\xf4h\xac\xa5\xf1\xc0\xa6\x87\xe7\xf3\x8c\x11\xec\x8d"
        + b"\x97j\x84_\xa31\x033\x11=\x10Y\xafG\x9c\xf0+1~\x06\\{\x89\xf8\x0b@"
        + b"\x97\xadf\xfc\x0f)\x84(\x13\xf1\xe7\x15\x99?e\xf6\xc7\x1a\x0cL'\xb5"
        + b"\xdfj\xfd9\xc6\x8f\xb7\xe1\x0eVMp\xbf\xad#\x84\xf6\x1b\xec\x06\xdf"
        + b"\x0f\xed\xb7\xfb;\xf1\xaf\xb3'\x1b6\x8f;\x0f\x90\xd4i\x00\x00\x00"
        + b"\x00IEND\xaeB`\x82"
    )



"""UniFi devices are network infrastructure.

Access points, gateways, power plugs, switches.
"""

from ..models.api import TypedApiResponse
from ..models.device import Device, DeviceListRequest, DeviceUpgradeRequest
from ..models.message import MessageKey
from .api_handlers import APIHandler


class Devices(APIHandler[Device]):
    """Represents network devices."""

    obj_id_key = "mac"
    item_cls = Device
    process_messages = (MessageKey.DEVICE,)
    api_request = DeviceListRequest.create()

    async def upgrade(self, mac: str) -> TypedApiResponse:
        """Upgrade network device."""
        return await self.controller.request(DeviceUpgradeRequest.create(mac))



"""UniFi devices are network infrastructure.

Access points, Gateways, Switches.
"""
from copy import deepcopy
from dataclasses import dataclass
from enum import IntEnum
import logging
from typing import Any, NotRequired, Self, TypedDict, cast

from .api import ApiItem, ApiRequest

LOGGER = logging.getLogger(__name__)


class TypedDeviceAntennaTable(TypedDict):
    """Device antenna table type definition."""

    default: bool
    id: int
    name: str
    wifi0_gain: int
    wifi1_gain: int


class TypedDeviceConfigNetwork(TypedDict):
    """Device config network type definition."""

    ip: str
    type: str


class TypedDeviceEthernetOverrides(TypedDict):
    """Device ethernet overrides type definition."""

    ifname: str
    networkgroup: str


class TypedDeviceEthernetTable(TypedDict):
    """Device ethernet table type definition."""

    mac: str
    name: str
    num_port: int


class TypedDeviceLastUplink(TypedDict):
    """Device last uplink type definition."""

    port_idx: int
    type: str
    uplink_mac: str
    uplink_device_name: str
    uplink_remote_port: int


class TypedDeviceLldpTable(TypedDict):
    """Device LLDP table type definition."""

    chassis_id: str
    chassis_id_subtype: str
    is_wired: bool
    local_port_idx: int
    local_port_name: str
    port_id: str


class TypedDeviceNetworkTable(TypedDict):
    """Device network table type definition."""

    _id: str
    attr_hidden_id: str
    attr_no_delete: bool
    dhcp_relay_enabled: bool
    dhcpd_dns_1: str
    dhcpd_dns_enabled: bool
    dhcpd_enabled: bool
    dhcpd_gateway_enabled: bool
    dhcpd_leasetime: int
    dhcpd_start: str
    dhcpd_stop: str
    dhcpd_time_offset_enabled: bool
    dhcpd_unifi_controller: str
    domain_name: str
    enabled: bool
    ip: str
    ip_subnet: str
    is_guest: bool
    is_nat: bool
    lte_lan_enabled: bool
    mac: str
    name: str
    networkgroup: str
    num_sta: int
    purpose: str
    rx_bytes: int
    rx_packets: int
    site_id: str
    tx_bytes: int
    tx_packets: int
    up: str
    vlan_enabled: bool


class TypedDeviceOutletOverrides(TypedDict, total=False):
    """Device outlet overrides type definition."""

    cycle_enabled: bool
    index: int
    has_relay: bool
    has_metering: bool
    name: str
    relay_state: bool


class TypedDeviceOutletTable(TypedDict):
    """Device outlet table type definition."""

    cycle_enabled: NotRequired[bool]
    index: int
    has_relay: NotRequired[bool]
    has_metering: NotRequired[bool]
    name: str
    outlet_caps: int
    outlet_voltage: NotRequired[str]
    outlet_current: NotRequired[str]
    outlet_power: NotRequired[str]
    outlet_power_factor: NotRequired[str]
    relay_state: bool


class TypedDevicePortOverrides(TypedDict, total=False):
    """Device port overrides type definition."""

    poe_mode: str
    port_idx: int
    portconf_id: str


class TypedDevicePortTableLldpTable(TypedDict):
    """Device port table mac table type definition."""

    lldp_chassis_id: str
    lldp_port_id: str
    lldp_system_name: str


class TypedDevicePortTableMacTable(TypedDict):
    """Device port table mac table type definition."""

    age: int
    mac: str
    static: bool
    uptime: int
    vlan: int


class TypedDevicePortTablePortDelta(TypedDict):
    """Device port table port delta type definition."""

    rx_bytes: int
    rx_packets: int
    time_delta: int
    time_delta_activity: int
    tx_bytes: int
    tx_packets: int


class TypedDevicePortTable(TypedDict):
    """Device port table type definition."""

    aggregated_by: bool
    attr_no_edit: bool
    autoneg: bool
    bytes_r: int
    dot1x_mode: str
    dot1x_status: str
    enable: bool
    flowctrl_rx: bool
    flowctrl_tx: bool
    full_duplex: bool
    ifname: NotRequired[str]
    is_uplink: bool
    jumbo: bool
    lldp_table: list[TypedDevicePortTableLldpTable]
    mac_table: list[TypedDevicePortTableMacTable]
    masked: bool
    media: str
    name: str
    op_mode: str
    poe_caps: int
    poe_class: str
    poe_current: str
    poe_enable: bool
    poe_good: bool
    poe_mode: str
    poe_power: str
    poe_voltage: str
    port_delta: TypedDevicePortTablePortDelta
    port_idx: NotRequired[int]
    port_poe: bool
    portconf_id: str
    rx_broadcast: int
    rx_bytes: int
    rx_bytes_r: int
    rx_dropped: int
    rx_errors: int
    rx_multicast: int
    rx_packets: int
    speed: int
    speed_caps: int
    stp_pathcost: int
    stp_state: str
    tx_broadcast: int
    tx_bytes: int
    tx_bytes_r: int
    tx_dropped: int
    tx_errors: int
    tx_multicast: int
    tx_packets: int
    up: NotRequired[bool]


class TypedDeviceRadioTable(TypedDict):
    """Device radio table type definition."""

    antenna_gain: int
    builtin_ant_gain: int
    builtin_antenna: bool
    channel: int
    current_antenna_gain: int
    hard_noise_floor_enabled: bool
    has_dfs: bool
    has_fccdfs: bool
    ht: str
    is_11ac: bool
    max_txpower: int
    min_rssi_enabled: bool
    min_txpower: int
    name: str
    nss: int
    radio: str
    radio_caps: int
    sens_level_enabled: bool
    tx_power_mode: str
    wlangroup_id: str


class TypedDeviceRadioTableStats(TypedDict):
    """Device radio table statistics type definition."""

    ast_be_xmit: int
    ast_cst: int
    ast_txto: int
    channel: int
    cu_self_rx: int
    cu_self_tx: int
    cu_total: int
    extchannel: int
    gain: int
    guest_num_sta: int
    name: str
    num_sta: int
    radio: str
    satisfaction: int
    state: str
    tx_packets: str
    tx_power: str
    tx_retries: str
    user_num_sta: str


class TypedDeviceSwitchCaps(TypedDict):
    """Device switch caps type definition."""

    feature_caps: int
    max_aggregate_sessions: int
    max_mirror_sessions: int
    vlan_caps: int


class TypedDeviceSysStats(TypedDict):
    """Device sys stats type definition."""

    loadavg_1: str
    loadavg_15: str
    loadavg_5: str
    mem_buffer: int
    mem_total: int
    mem_used: int


class TypedDeviceSystemStats(TypedDict):
    """Device system stats type definition."""

    cpu: str
    mem: str
    uptime: str


class TypedDeviceUplink(TypedDict):
    """Device uplink type definition."""

    full_duplex: bool
    ip: str
    mac: str
    max_speed: int
    max_vlan: int
    media: str
    name: str
    netmask: str
    num_port: int
    rx_bytes: int
    rx_bytes_r: int
    rx_dropped: int
    rx_errors: int
    rx_multicast: int
    rx_packets: int
    speed: int
    tx_bytes: int
    tx_bytes_r: int
    tx_dropped: int
    tx_errors: int
    tx_packets: int
    type: str
    up: bool
    uplink_mac: str
    uplink_remote_port: int


class TypedDeviceWlanOverrides(TypedDict):
    """Device wlan overrides type definition."""

    name: str
    radio: str
    radio_name: str
    wlan_id: str


class TypedDeviceSpeedtestStatus(TypedDict):
    """Device speedtest status type definition."""

    latency: int
    rundate: int
    runtime: int
    status_download: int
    status_ping: int
    status_summary: int
    status_upload: int
    xput_download: float
    xput_upload: float


class TypedDevice(TypedDict):
    """Device type definition."""

    _id: str
    _uptime: int
    adoptable_when_upgraded: bool
    adopted: bool
    antenna_table: list[TypedDeviceAntennaTable]
    architecture: str
    adoption_completed: int
    board_rev: NotRequired[int]
    bytes: int
    bytes_d: int
    bytes_r: int
    cfgversion: int
    config_network: TypedDeviceConfigNetwork
    connect_request_ip: str
    connect_request_port: str
    considered_lost_at: int
    country_code: int
    countrycode_table: list  # type: ignore[type-arg]
    device_id: str
    dhcp_server_table: list  # type: ignore[type-arg]
    disabled: NotRequired[bool]
    disconnection_reason: str
    displayable_version: str
    dot1x_portctrl_enabled: bool
    downlink_table: list  # type: ignore[type-arg]
    element_ap_serial: str
    element_peer_mac: str
    element_uplink_ap_mac: str
    ethernet_overrides: list[TypedDeviceEthernetOverrides]
    ethernet_table: list[TypedDeviceEthernetTable]
    fan_level: int
    flowctrl_enabled: bool
    fw_caps: int
    gateway_mac: str
    general_temperature: NotRequired[int]
    guest_num_sta: int
    guest_wlan_num_sta: int
    guest_token: str
    has_eth1: bool
    has_fan: bool
    has_speaker: bool
    has_temperature: bool
    hash_id: str
    hide_ch_width: str
    hw_caps: int
    inform_ip: str
    inform_url: str
    internet: bool
    ip: NotRequired[str]
    isolated: bool
    jumboframe_enabled: bool
    kernel_version: str
    known_cfgversion: str
    last_seen: int
    last_uplink: TypedDeviceLastUplink
    lcm_brightness: int
    lcm_brightness_override: bool
    lcm_idle_timeout_override: bool
    lcm_night_mode_begins: str
    lcm_night_mode_enabled: bool
    lcm_night_mode_ends: str
    lcm_tracker_enabled: bool
    led_override: str
    led_override_color: str
    led_override_color_brightnes: int
    license_state: str
    lldp_table: list[TypedDeviceLldpTable]
    locating: bool
    mac: str
    manufacturer_id: int
    meshv3_peer_mac: str
    model: str
    model_in_eol: bool
    model_in_lts: bool
    model_incompatible: bool
    name: str
    network_table: list[TypedDeviceNetworkTable]
    next_heartbeat_at: int
    next_interval: int
    num_desktop: int
    num_handheld: int
    num_mobile: int
    num_sta: int
    outdoor_mode_override: str
    outlet_ac_power_budget: str
    outlet_ac_power_consumption: str
    outlet_enabled: bool
    outlet_overrides: list[TypedDeviceOutletOverrides]
    outlet_table: list[TypedDeviceOutletTable]
    overheating: bool
    power_source_ctrl_enabled: bool
    prev_non_busy_state: int
    provisioned_at: int
    port_overrides: list[TypedDevicePortOverrides]
    port_table: NotRequired[list[TypedDevicePortTable]]
    radio_table: list[TypedDeviceRadioTable]
    radio_table_stats: list[TypedDeviceRadioTableStats]
    required_version: str
    rollupgrade: bool
    rx_bytes: int
    rx_bytes_d: int
    satisfaction: int
    scan_radio_table: list  # type: ignore[type-arg]
    scanning: bool
    serial: str
    site_id: str
    spectrum_scanning: bool
    speedtest_status: TypedDeviceSpeedtestStatus | None
    ssh_session_table: list  # type: ignore[type-arg]
    start_connected_millis: int
    start_disconnected_millis: int
    stat: dict  # type: ignore[type-arg]
    state: int
    stp_priority: str
    stp_version: str
    switch_caps: TypedDeviceSwitchCaps
    sys_error_caps: int
    sys_stats: TypedDeviceSysStats
    syslog_key: str
    system_stats: TypedDeviceSystemStats
    two_phase_adopt: bool
    tx_bytes: int
    tx_bytes_d: int
    type: str
    unsupported: bool
    unsupported_reason: int
    upgradable: bool
    upgrade_state: int
    upgrade_to_firmware: str
    uplink: TypedDeviceUplink
    uplink_depth: int
    uplink_table: list  # type: ignore[type-arg]
    uptime: int
    user_num_sta: int
    user_wlan_num_sta: int
    usg_caps: int
    vap_table: list[dict]  # type: ignore[type-arg]
    version: str
    vwireEnabled: bool
    vwire_table: list  # type: ignore[type-arg]
    vwire_vap_table: list  # type: ignore[type-arg]
    wifi_caps: int
    wlan_overrides: list[TypedDeviceWlanOverrides]
    wlangroup_id_na: str
    wlangroup_id_ng: str
    x_aes_gcm: bool
    x_authkey: str
    x_fingerprint: str
    x_has_ssh_hostkey: bool
    x_inform_authkey: str
    x_ssh_hostkey_fingerprint: str
    x_vwirekey: str


class DeviceState(IntEnum):
    """Enum for device states."""

    DISCONNECTED = 0
    CONNECTED = 1
    PENDING = 2
    FIRMWARE_MISMATCH = 3
    UPGRADING = 4
    PROVISIONING = 5
    HEARTBEAT_MISSED = 6
    ADOPTING = 7
    DELETING = 8
    INFORM_ERROR = 9
    ADOPTION_FALIED = 10
    ISOLATED = 11

    UNKNOWN = -1

    @classmethod
    def _missing_(cls, value: object) -> "DeviceState":
        """Set default enum member if an unknown value is provided."""
        LOGGER.warning("Unsupported device state %s %s", value, cls)
        return DeviceState.UNKNOWN


@dataclass
class DeviceListRequest(ApiRequest):
    """Request object for device list."""

    @classmethod
    def create(cls) -> Self:
        """Create device list request."""
        return cls(method="get", path="/stat/device")


@dataclass
class DevicePowerCyclePortRequest(ApiRequest):
    """Request object for power cycle PoE port."""

    @classmethod
    def create(cls, mac: str, port_idx: int) -> Self:
        """Create power cycle of PoE request."""
        return cls(
            method="post",
            path="/cmd/devmgr",
            data={
                "cmd": "power-cycle",
                "mac": mac,
                "port_idx": port_idx,
            },
        )


@dataclass
class DeviceRestartRequest(ApiRequest):
    """Request object for device restart."""

    @classmethod
    def create(cls, mac: str, soft: bool = True) -> Self:
        """Create device restart request.

        Hard is specifically for PoE switches and will additionally cycle PoE ports.
        """
        return cls(
            method="post",
            path="/cmd/devmgr",
            data={
                "cmd": "restart",
                "mac": mac,
                "reboot_type": "soft" if soft else "hard",
            },
        )


@dataclass
class DeviceUpgradeRequest(ApiRequest):
    """Request object for device upgrade."""

    @classmethod
    def create(cls, mac: str) -> Self:
        """Create device upgrade request."""
        return cls(
            method="post",
            path="/cmd/devmgr",
            data={
                "cmd": "upgrade",
                "mac": mac,
            },
        )


@dataclass
class DeviceSetOutletRelayRequest(ApiRequest):
    """Request object for outlet relay state."""

    @classmethod
    def create(cls, device: "Device", outlet_idx: int, state: bool) -> Self:
        """Create device outlet relay state request.

        True:  outlet power output on.
        False: outlet power output off.
        """
        existing_override = False
        for outlet_override in device.outlet_overrides:
            if outlet_idx == outlet_override["index"]:
                outlet_override["relay_state"] = state
                existing_override = True
                break

        if not existing_override:
            name = device.outlet_table[outlet_idx - 1].get("name", "")
            device.outlet_overrides.append(
                {
                    "index": outlet_idx,
                    "name": name,
                    "relay_state": state,
                }
            )

        return cls(
            method="put",
            path=f"/rest/device/{device.id}",
            data={"outlet_overrides": device.outlet_overrides},
        )


@dataclass
class DeviceSetOutletCycleEnabledRequest(ApiRequest):
    """Request object for outlet cycle_enabled flag."""

    @classmethod
    def create(cls, device: "Device", outlet_idx: int, state: bool) -> Self:
        """Create device outlet outlet cycle_enabled flag request.

        True:  UniFi Network will power cycle this outlet if the internet goes down.
        False: UniFi Network will not power cycle this outlet if the internet goes down.
        """
        existing_override = False
        for outlet_override in device.outlet_overrides:
            if outlet_idx == outlet_override["index"]:
                outlet_override["cycle_enabled"] = state
                existing_override = True
                break

        if not existing_override:
            name = device.outlet_table[outlet_idx - 1].get("name", "")
            device.outlet_overrides.append(
                {
                    "index": outlet_idx,
                    "name": name,
                    "cycle_enabled": state,
                }
            )

        return cls(
            method="put",
            path=f"/rest/device/{device.id}",
            data={"outlet_overrides": device.outlet_overrides},
        )


@dataclass
class DeviceSetPoePortModeRequest(ApiRequest):
    """Request object for setting port PoE mode."""

    @classmethod
    def create(
        cls,
        device: "Device",
        port_idx: int | None = None,
        mode: str | None = None,
        targets: list[tuple[int, str]] | None = None,
    ) -> Self:
        """Create device set port PoE mode request.

        Auto, 24v, passthrough, off.
        Make sure to not overwrite any existing configs.
        """
        overrides: list[tuple[int, str]] = []
        if port_idx is not None and mode is not None:
            overrides.append((port_idx, mode))
        elif targets is not None:
            overrides = targets
        else:
            raise AttributeError

        port_overrides = deepcopy(device.port_overrides)

        for override in overrides:
            port_idx, mode = override

            existing_override = False
            for port_override in port_overrides:
                if port_idx == port_override.get("port_idx"):
                    port_override["poe_mode"] = mode
                    existing_override = True
                    break

            if existing_override:
                continue

            port_override = {"port_idx": port_idx, "poe_mode": mode}
            if portconf_id := device.port_table[port_idx - 1].get("portconf_id"):
                port_override["portconf_id"] = portconf_id
            port_overrides.append(port_override)

        return cls(
            method="put",
            path=f"/rest/device/{device.id}",
            data={"port_overrides": port_overrides},
        )


class Device(ApiItem):
    """Represents a network device."""

    raw: TypedDevice

    @property
    def board_revision(self) -> int:
        """Board revision of device."""
        return self.raw.get("board_rev", 0)

    @property
    def considered_lost_at(self) -> int:
        """When device is considered lost."""
        return self.raw["considered_lost_at"]

    @property
    def disabled(self) -> bool:
        """Is device disabled."""
        return self.raw.get("disabled", False)

    @property
    def downlink_table(self) -> list[dict[str, Any]]:
        """All devices with device as uplink."""
        return self.raw.get("downlink_table", [])

    @property
    def fan_level(self) -> int | None:
        """Fan level of device."""
        return self.raw.get("fan_level")

    @property
    def general_temperature(self) -> int | None:
        """General temperature of device."""
        return self.raw.get("general_temperature")

    @property
    def has_fan(self) -> bool:
        """Do device have a fan."""
        return self.raw.get("has_fan", False)

    @property
    def has_temperature(self) -> bool:
        """Do the device have a general temperature."""
        return self.raw.get("has_temperature", False)

    @property
    def id(self) -> str:
        """ID of device."""
        return self.raw["device_id"]

    @property
    def ip(self) -> str | None:
        """IP of device."""
        return self.raw.get("ip")

    @property
    def last_seen(self) -> int | None:
        """When was device last seen."""
        return self.raw.get("last_seen")

    @property
    def lldp_table(self) -> list[TypedDeviceLldpTable]:
        """All clients and devices directly attached to device."""
        return self.raw.get("lldp_table", [])

    @property
    def mac(self) -> str:
        """MAC address of device."""
        return self.raw["mac"]

    @property
    def model(self) -> str:
        """Model of device."""
        return self.raw["model"]

    @property
    def name(self) -> str:
        """Name of device."""
        return self.raw.get("name", "")

    @property
    def next_heartbeat_at(self) -> int | None:
        """Next heart beat full UNIX time."""
        return self.raw.get("next_heartbeat_at")

    @property
    def next_interval(self) -> int:
        """Next heart beat in seconds."""
        return self.raw.get("next_interval", 30)

    @property
    def overheating(self) -> bool:
        """Is device overheating."""
        return self.raw.get("overheating", False)

    @property
    def outlet_overrides(self) -> list[TypedDeviceOutletOverrides]:
        """Overridden outlet configuration."""
        return self.raw.get("outlet_overrides", [])

    @property
    def outlet_ac_power_budget(self) -> str | None:
        """The amount of power available to outlets."""
        return self.raw.get("outlet_ac_power_budget")

    @property
    def outlet_ac_power_consumption(self) -> str | None:
        """The amount of power consumed by all outlets."""
        return self.raw.get("outlet_ac_power_consumption")

    @property
    def outlet_table(self) -> list[TypedDeviceOutletTable]:
        """List of outlets."""
        return self.raw.get("outlet_table", [])

    @property
    def port_overrides(self) -> list[TypedDevicePortOverrides]:
        """Overridden port configuration."""
        return self.raw.get("port_overrides", [])

    @property
    def port_table(self) -> list[TypedDevicePortTable]:
        """List of ports and data."""
        return self.raw.get("port_table", [])

    @property
    def speedtest_status(self) -> TypedDeviceSpeedtestStatus | None:
        """Speedtest status."""
        if value := self.raw.get("speedtest-status"):
            return cast(TypedDeviceSpeedtestStatus, value)
        return None

    @property
    def state(self) -> DeviceState:
        """State of device."""
        return DeviceState(self.raw["state"])

    @property
    def sys_stats(self) -> TypedDeviceSysStats:
        """Output from top."""
        return self.raw["sys_stats"]

    @property
    def system_stats(self) -> tuple[str, str, str]:
        """System statistics."""
        data = self.raw["system-stats"]  # type: ignore [typeddict-item]
        return (data.get("cpu", ""), data.get("mem", ""), data.get("uptime", ""))

    @property
    def type(self) -> str:
        """Type of device."""
        return self.raw["type"]

    @property
    def version(self) -> str:
        """Firmware version."""
        return self.raw["version"]

    @property
    def upgradable(self) -> bool:
        """Is a new firmware available."""
        return self.raw.get("upgradable", False)

    @property
    def upgrade_to_firmware(self) -> str:
        """Firmware version to update to."""
        return self.raw.get("upgrade_to_firmware", "")

    @property
    def uplink(self) -> TypedDeviceUplink:
        """Information about uplink."""
        return self.raw["uplink"]

    @property
    def uplink_depth(self) -> int | None:
        """Hops to gateway."""
        return self.raw.get("uplink_depth")

    @property
    def uptime(self) -> int:
        """Uptime of device."""
        return self.raw.get("uptime", 0)

    @property
    def user_num_sta(self) -> int:
        """Amount of connected clients."""
        value = self.raw.get("user-num_sta")
        assert isinstance(value, int)
        return value

    @property
    def wlan_overrides(self) -> list[TypedDeviceWlanOverrides]:
        """Wlan configuration override."""
        return self.raw.get("wlan_overrides", [])

    def __repr__(self) -> str:
        """Return the representation."""
        return f"<Device {self.name}: {self.mac}>"
